# vim: ts=4:sts=4:sw=4
#
# @date 2019-09-30
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
https://earth.esa.int/web/sentinel/technical-guides/sentinel-3-slstr/level-1/vis-and-swir-radiances
https://sentinel.esa.int/web/sentinel/user-guides/sentinel-3-slstr/data-formats/level-1
https://earth.esa.int/web/sentinel/user-guides/sentinel-3-slstr/product-types/level-1b
https://sentinel.esa.int/documents/247904/685236/Sentinel-3_User_Handbook#%5B%7B%22num%22%3A235%2C%22gen%22%3A0%7D%2C%7B%22name%22%3A%22XYZ%22%7D%2C54%2C628%2C0%5D

S1 555   nm            B
S2 659   nm            G
S3 865   nm            R
S4 1.375 microm
S5 1.610 microm
S6 2.25  microm
S7 3.74  microm        IR
S8 10.85 microm        IR
S9 12    microm        IR
"""

import os
import numpy
import numpy.typing
import typing
import netCDF4
import logging
import datetime
import scipy.ndimage
import defusedxml.ElementTree as ET
import idf_converter.lib
from idf_converter.lib.types import InputOptions, OutputOptions
from idf_converter.lib.types import ReaderResult, FormatterJob

logger = logging.getLogger(__name__)

N2SSTInputs = typing.Tuple[typing.Dict[str, typing.Any], numpy.typing.NDArray]
CloudBitsToMask002 = typing.Tuple[str, str, str, str, str, str, str]
CloudBitsToMask003 = typing.Tuple[str, str, str, str, str, str, str, str, str,
                                  str, str, str]
DATA_MODEL = 'SWATH'


class InputPathMissing(Exception):
    """Error raised when the "path" input option has not been specified by the
    user."""
    pass


class GeoCoordinatesNotFound(IOError):
    """Error raised when the path for the geographical coordinates file does
    not exist."""
    def __init__(self, path: str) -> None:
        """"""
        self.path = path


class BrightnessTemperatureFileNotFound(IOError):
    """Error raised when the path for the file which contains brightness
    temperature values does not exist."""
    def __init__(self, path: str) -> None:
        """"""
        self.path = path


class SSTCoefficientsPathNotFound(IOError):
    """Error raised when the path for the SST coefficients file does not exist.
    """
    def __init__(self, path: str) -> None:
        """"""
        self.path = path


class BaselineNotSupported(Exception):
    """Error raised when the input file corresponds to a product version which
    is either not supported or has not been properly tested with this reader.
    """
    def __init__(self, baseline: str) -> None:
        """"""
        self.baseline = baseline


def help() -> typing.Tuple[str, str]:
    """Describe options supported by this reader.

    Returns
    -------
    str
        Description of the supported options
    """
    inp = ('    path\tPath of the input file',)
    out = ('',)
    return ('\n'.join(inp), '\n'.join(out))


def read_manifest(input_dir: str) -> typing.Dict[str, typing.Any]:
    """ Read metadata in xml file. Return start and track Offset, number of
    lines and elements spatial resolution for Tie Points, 1 km nadir and 1 km
    oblique grid."""
    nfile = os.path.join(input_dir, 'xfdumanifest.xml')
    tree = ET.parse(nfile)
    root = tree.getroot()
    param = {}
    s3path = '{http://www.esa.int/safe/sentinel/sentinel-3/1.0}'
    slstrpath = '{http://www.esa.int/safe/sentinel/sentinel-3/slstr/1.0}'
    for a_child in root.findall("./*/metadataObject/metadataWrap/xmlData/*"):
        for child in a_child:
            if child.tag == '{}nadirImageSize'.format(slstrpath):
                if child.attrib['grid'] == '1 km':
                    for item in child.findall("./*"):
                        if item.tag == '{}startOffset'.format(s3path):
                            _item = item.text.encode('utf8')
                            param['startOffset_n'] = float(_item)
                        if item.tag == '{}trackOffset'.format(s3path):
                            _item = item.text.encode('utf8')
                            param['trackOffset_n'] = float(_item)
                        if item.tag == '{}rows'.format(s3path):
                            _item = item.text.encode('utf8')
                            param['line_n'] = int(_item)
                        if item.tag == '{}columns'.format(s3path):
                            _item = item.text.encode('utf8')
                            param['elem_n'] = int(_item)
                if child.attrib['grid'] == 'Tie Points':
                    for item in child.findall("./*"):
                        if item.tag == '{}startOffset'.format(s3path):
                            _item = item.text.encode('utf8')
                            param['startOffset_tx'] = float(_item)
                        if item.tag == '{}trackOffset'.format(s3path):
                            _item = item.text.encode('utf8')
                            param['trackOffset_tx'] = float(_item)
                        if item.tag == '{}rows'.format(s3path):
                            param['line_tx'] = int(item.text.encode('utf8'))
                        if item.tag == '{}columns'.format(s3path):
                            param['elem_tx'] = int(item.text.encode('utf8'))
            if child.tag == '{}obliqueImageSize'.format(slstrpath):
                if child.attrib['grid'] == '1 km':
                    for item in child.findall("./*"):
                        if item.tag == '{}startOffset'.format(s3path):
                            _item = item.text.encode('utf8')
                            param['startOffset_o'] = float(_item)
                        if item.tag == '{}trackOffset'.format(s3path):
                            _item = item.text.encode('utf8')
                            param['trackOffset_o'] = float(_item)
                        if item.tag == '{}rows'.format(s3path):
                            param['line_o'] = int(item.text.encode('utf8'))
                        if item.tag == '{}columns'.format(s3path):
                            param['elem_o'] = int(item.text.encode('utf8'))
            if child.tag == '{}resolution'.format(slstrpath):
                if child.attrib['grid'] == '1 km':
                    for item in child.findall("./*"):
                        if item.tag == '{}spatialResolution'.format(slstrpath):
                            param['res_n'] = float(item.text.encode('utf8'))
                if child.attrib['grid'] == 'Tie Points':
                    for item in child.findall("./*"):
                        if item.tag == '{}spatialResolution'.format(slstrpath):
                            param['res_tx'] = float(item.text.encode('utf8'))
            if child.tag == f'{s3path}baselineCollection':
                param['baselineCollection'] = child.text.encode('utf-8')
    return param


def get_n2_sst_inputs(input_dir: str, coeff_path: str) -> N2SSTInputs:
    """"""
    param = read_manifest(input_dir)

    quality_flags_path = os.path.join(input_dir, 'flags_in.nc')
    quality_handler = netCDF4.Dataset(quality_flags_path, 'r')
    quality_flags = idf_converter.lib.extract_variable_values(quality_handler,
                                                              'confidence_in')
    cloud_flags = idf_converter.lib.extract_variable_values(quality_handler,
                                                            'cloud_in')
    quality_handler.close()

    # INFRARED MASK
    bits = {'coastline': 0,
            'ocean': 1,
            'tidal': 2,
            'land': 3,
            'inland_water': 4,
            'unfilled': 5,
            'spare': 6,
            'spare_': 7,
            'cosmetic': 8,
            'duplicate': 9,
            'day': 10,
            'twilight': 11,
            'sun_glint': 12,
            'snow': 13,
            'summary_cloud': 14,
            'summary_pointing': 15}

    # Invalidate if these flags are all off
    on_flags = numpy.ushort(0)
    on_flags = on_flags + numpy.ushort(1 << bits['day'])

    # Invalidate if any of these flags is on
    off_flags = numpy.ushort(0)
    off_flags = off_flags + numpy.ushort(1 << bits['coastline'])
    off_flags = off_flags + numpy.ushort(1 << bits['land'])
    off_flags = off_flags + numpy.ushort(1 << bits['snow'])

    ir_mask = (numpy.bitwise_and(quality_flags, off_flags) > 0)

    # CLOUD MASK
    cloud_bits = {'visible': 0,
                  'threshold': 1,
                  'small_histogram1.6': 2,
                  'large_histogram1.6': 3,
                  'small_histogram2.25': 4,
                  'large_histogram2.25': 5,
                  'spatial_coherence': 6,
                  'gross_cloud': 7,
                  'thin_cirrus': 8,
                  'medium_high': 9,
                  'fog_low_stratus': 10,
                  '11_12_view_difference': 11,
                  '3.7_11_view_difference': 12,
                  'thermal_histogram': 13}

    collection = param['baselineCollection']
    cloud_bits_to_mask: typing.Union[CloudBitsToMask002, CloudBitsToMask003]
    if collection == b'002':
        cloud_bits_to_mask = ('visible',
                              'threshold',
                              # 'small_histogram1.6',
                              # 'small_histogram2.25',
                              'spatial_coherence',
                              'gross_cloud',
                              'thin_cirrus',
                              'medium_high',
                              'fog_low_stratus',
                              # '11_12_view_difference',
                              )
    elif collection in (b'003', b'004'):
        cloud_bits_to_mask = ('visible',
                              'threshold',
                              'small_histogram1.6',
                              'large_histogram1.6',
                              'small_histogram2.25',
                              'large_histogram2.25',
                              'spatial_coherence',
                              'gross_cloud',
                              'thin_cirrus',
                              'medium_high',
                              'fog_low_stratus',
                              '11_12_view_difference',
                              # '3.7_11_view_difference',
                              )
    else:
        logger.error('Baseline collection should be 002, 003 or 004')
        raise BaselineNotSupported(collection)

    mask_ref = numpy.ushort(0)
    mask_spatial = numpy.ushort(0)
    # Enable selected mask bits
    for mask_name in cloud_bits_to_mask:
        if mask_name == 'spatial_coherence':
            mask_spatial = mask_spatial + (1 << cloud_bits[mask_name])
        else:
            mask_ref = mask_ref + (1 << cloud_bits[mask_name])

    # Build cloud mask
    _cloud_mask = numpy.zeros(shape=cloud_flags.shape, dtype='bool')
    _cloud_mask = (_cloud_mask |
                   (numpy.bitwise_and(cloud_flags, mask_spatial) > 1) |
                   (numpy.bitwise_and(cloud_flags, mask_ref) > 1))

    binary_dilation = scipy.ndimage.binary_dilation

    # Dilate non-masked data
    dil_kern = numpy.ones((9, 9), dtype='bool')
    cloud_mask_dil = ~(binary_dilation(~_cloud_mask, structure=dil_kern))

    # Erode cloud mask
    dil_kern = numpy.ones((9, 9), dtype='bool')
    cloud_mask = (binary_dilation(cloud_mask_dil, structure=dil_kern))
    cloud_mask = (cloud_mask |
                  (numpy.bitwise_and(cloud_flags, mask_ref) > 1))

    # -------------------------------------------------------------------------
    # Do not mask data at night or twilight for ir channels.
    data_mask = numpy.zeros(quality_flags.shape, dtype='bool')
    data_mask = (data_mask | cloud_mask | ir_mask)

    # - Get satellite angles
    nfile = os.path.join(input_dir, 'geometry_tn.nc')
    fid = netCDF4.Dataset(nfile, 'r')
    sat_zen_n = idf_converter.lib.extract_variable_values(fid, 'sat_zenith_tn')
    fid.close()

    # - Get TCWV from MET file
    nfile = os.path.join(input_dir, 'met_tx.nc')
    fid = netCDF4.Dataset(nfile, 'r')
    tcwv_var_id = 'total_column_water_vapour_tx'
    _tcwv = idf_converter.lib.extract_variable_values(fid, tcwv_var_id)
    tcwv = _tcwv[0, :, :]
    fid.close()

    # -- Read SST coefficents for nadir
    fid = netCDF4.Dataset(coeff_path, 'r')
    coeffs_n2 = idf_converter.lib.extract_variable_values(fid,
                                                          'coefficients_LUT')
    fid.close()

    n2_sst_inputs = {'tcwv': tcwv,
                     'sat_zen_n': sat_zen_n,
                     'param': param,
                     'coeffs_n2': coeffs_n2,
                     'n11_id': 'S8_BT_in',
                     'n12_id': 'S9_BT_in'}
    return n2_sst_inputs, data_mask


def get_sst_mask(variables: typing.Dict[str, typing.Dict[str, typing.Any]],
                 var_id: str
                 ) -> numpy.typing.NDArray:
    """"""
    n11_mask = numpy.ma.getmaskarray(variables['S8_BT_in']['array'])
    n12_mask = numpy.ma.getmaskarray(variables['S9_BT_in']['array'])
    result: numpy.typing.NDArray = (n11_mask | n12_mask)
    return result


def read_data(input_opts: InputOptions,
              output_opts: OutputOptions
              ) -> typing.Iterator[ReaderResult]:
    """Read input file, extract data and metadata, store them in a Granule
    object then prepare formatting instructions to finalize the conversion to
    IDF format.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule, list)
        A tuple which contains four elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the extracted information
          has been stored

        - a :obj:list of :obj:dict describing the formatting operations that
          the converter must perform before serializing the result in IDF
          format
    """

    variables = {'S7_BT_in': {'name': 'S7',
                              'valid_min': 220.000,
                              'valid_max': 310.000,
                              'options': {}},
                 'S8_BT_in': {'name': 'S8',
                              'valid_min': 197.400,
                              'valid_max': 306.000,
                              'options': {}},
                 'S9_BT_in': {'name': 'S9',
                              'valid_min': 190.150,
                              'valid_max': 306.150,
                              'options': {}},
                 'sst': {'name': 'sea_surface_temperature',
                         'long_name': 'Sea surface temperature',
                         'units': 'K',
                         'valid_min': 270.15,
                         'valid_max': 310.15,
                         'options': {}}}

    radiance_vars = ['S7_BT_in', 'S8_BT_in', 'S9_BT_in']

    idf_version = output_opts.get('idf_version', '1.0')
    granule = idf_converter.lib.create_granule(idf_version, DATA_MODEL)
    granule.vars = variables

    _input_dir = input_opts.get('path', None)
    if _input_dir is None:
        raise InputPathMissing()
    input_dir = os.path.normpath(_input_dir)

    sst_coeffs_path = input_opts.get('sst_coeffs_path', None)
    if sst_coeffs_path is None:
        del variables['sst']
    elif not os.path.exists(sst_coeffs_path):
        raise SSTCoefficientsPathNotFound(sst_coeffs_path)

    # Check existance of required files
    geo_path = os.path.join(input_dir, 'geodetic_in.nc')
    if not os.path.exists(geo_path):
        raise GeoCoordinatesNotFound(geo_path)

    # Read variables
    for var_id in radiance_vars:
        file_name = f'{var_id}.nc'
        bt_path = os.path.join(input_dir, file_name)
        if not os.path.exists(bt_path):
            raise BrightnessTemperatureFileNotFound(bt_path)

        f_handler = netCDF4.Dataset(bt_path, 'r')
        idf_converter.lib.extract_variable_attributes(f_handler, var_id,
                                                      granule)
        band = idf_converter.lib.extract_variable_values(f_handler, var_id)
        granule.vars[var_id]['array'] = numpy.ma.array(band)
        f_handler.close()

        band_mask = numpy.ma.getmaskarray(variables[var_id]['array'])
        if numpy.all(band_mask):
            logger.warning(f'All values masked for channel {var_id}')

    # Extract geo coordinates information
    lat_varname = 'latitude_in'
    lon_varname = 'longitude_in'
    geo_handler = netCDF4.Dataset(geo_path, 'r')
    idf_converter.lib.extract_global_attributes(geo_handler, input_opts,
                                                granule)
    nrow = geo_handler.dimensions['rows'].size
    ncell = geo_handler.dimensions['columns'].size
    lon = idf_converter.lib.extract_variable_values(geo_handler, lon_varname)
    lon = numpy.ma.array(lon)
    lat = idf_converter.lib.extract_variable_values(geo_handler, lat_varname)
    lat = numpy.ma.array(lat)
    start_time_str = geo_handler.start_time
    stop_time_str = geo_handler.stop_time
    geo_handler.close()

    granule.dims['row'] = nrow
    granule.dims['cell'] = ncell

    # Handle longitude continuity
    dlon = lon[1:, :] - lon[:-1, :]
    if 180.0 <= numpy.max(numpy.abs(dlon)):
        lon0 = lon[0, 0] + 180.0
        lon[:, :] = numpy.mod(lon[:, :] - lon0, 360.0) + lon0

    granule.vars['lat'] = {'array': lat,
                           'units': 'degrees_north',
                           'datatype': lat.dtype,
                           'options': {}}
    granule.vars['lon'] = {'array': lon,
                           'units': 'degrees_east',
                           'datatype': lon.dtype,
                           'options': {}}

    # Format time information
    time_fmt = '%Y-%m-%dT%H:%M:%S.%fZ'
    start_dt = datetime.datetime.strptime(start_time_str, time_fmt)
    stop_dt = datetime.datetime.strptime(stop_time_str, time_fmt)

    granule_name = os.path.basename(input_dir)

    granule.meta['idf_granule_id'] = granule_name
    granule.meta['idf_subsampling_factor'] = 0
    granule.meta['idf_spatial_resolution'] = 1000  # 1km
    granule.meta['idf_spatial_resolution_units'] = 'm'
    granule.meta['time_coverage_start'] = start_dt
    granule.meta['time_coverage_end'] = stop_dt
    granule.meta['institution'] = 'ESA'
    granule.meta['platform'] = 'Sentinel-3'
    granule.meta['platform_type'] = 'leo satellite'
    granule.meta['sensor'] = 'SLSTR'
    granule.meta['sensor_type'] = 'dual scan temperature radiometer'

    # Add transforms
    transforms = []

    channels = ['S7_BT_in', 'S8_BT_in', 'S9_BT_in']
    if sst_coeffs_path:
        n2_sst_inputs, data_mask = get_n2_sst_inputs(input_dir,
                                                     sst_coeffs_path)
        transforms.append(('n2_sst_from_slstr', {'input_data': n2_sst_inputs}))

        # mask night data for rgb and invalid data for ir (cloud, land,
        # range value).
        night_mask_job: FormatterJob = ('static_common_mask',
                                        {'targets': ['sst'],
                                         'mask': data_mask})
        transforms.append(night_mask_job)

        sst_dyn_mask = {'sst': get_sst_mask}
        sst_input_mask_job: FormatterJob = ('mask_methods',
                                            {'targets': ['sst'],
                                             'methods': sst_dyn_mask})
        transforms.append(sst_input_mask_job)

        channels.append('sst')

    output_opts['__export'] = channels

    yield input_opts, output_opts, granule, transforms



# vim: ts=4:sts=4:sw=4
#
# @date 2019-10-10
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import os
import numpy
import numpy.typing
import typing
import netCDF4
import logging
import idf_converter.lib
import idf_converter.lib.time
import idf_converter.lib.generic_grid
from idf_converter.lib.types import InputOptions, OutputOptions
from idf_converter.lib.types import Granule, ReaderResult

logger = logging.getLogger(__name__)

DATA_MODEL = 'GRID_LATLON'


class MissingLongitudeDimension(Exception):
    """Error raised when the "lon_dimension" input option has not been
    specified by the user. If the "lon_variable" has been provided, the reader
    will try this value too, in case the dimension and the variable share the
    same identifier."""
    pass


class MissingLatitudeDimension(Exception):
    """Error raised when the "lat_dimension" input option has not been
    specified by the user. If the "lat_variable" has been provided, the reader
    will try this value too, in case the dimension and the variable share the
    same identifier."""
    pass


class MissingVariablesList(Exception):
    """Error raised when the "variables" input option has not been specified
    by the user."""
    pass


class MissingValidTime(Exception):
    """Error raised when time values are read from a variable in the input
    file but none of them is valid (masked values)."""
    def __init__(self, time_var_id: str) -> None:
        """"""
        self.time_var_id = time_var_id


class InputPathMissing(Exception):
    """Error raised when the "path" input option has not been specified by the
    user."""
    pass


def help() -> typing.Tuple[str, str]:
    """Describe options supported by this reader.

    Returns
    -------
    str
        Description of the supported options
    """
    inp = ('    path\tPath of the input file',
           '    variables\tComma-separated list of variable identifiers to'
           ' extract from the input file',
           '    lon_variable\tIdentifier of the longitude variable',
           '    lat_variable\tIdentifier of the latitude variable',
           '    lon_dimension\tIdentifier of the longitude dimension'
           ' (defaults to the value of lon_variable)',
           '    lat_dimension\tIdentifier of the latitude dimension'
           ' (defaults to the value of lat_variable)',
           '    lon_reversed\tSet to yes|true|1 if longitude values should'
           ' be traversed in reverse order (defaults to no|false|0)',
           '    lat_reversed\tSet to yes|true|1 if latitude values should'
           ' be traversed in reverse order (defaults to no|false|0)',
           '    time_variable\tIdentifier of the time variable (optional).'
           ' If time_variable is not defined, then time_coverage_start'
           ' and time_coverage_end attributes must be defined either in the'
           ' input file or via global_overrides',
           '    depth_variable\tIdentifier of the depth variable (optional).'
           ' If depth_variable is defined, it will be assumed that it'
           ' corresponds to the slowest changed dimension after the time '
           ' dimension',
           '    global_overrides\tDefinition of global attributes that are'
           ' either missing or erroneous in the input file. The value of this'
           ' option is a comma-separated list of key:value',
           '    variable_overrides_VAR\tDefinition of variable attributes'
           ' that are either missing or erroneous in the input file.'
           ' VAR is the identifier of the variable and the value of this'
           ' option is a comma-separated list of key:value',
           '    time_coverage_relative_start\tOffset (in seconds) that must'
           ' be applied to the values of the time variable to obtain the lower'
           ' bound of the time coverage. This option is required when there is'
           ' more than one time value in the input file or when it is not'
           ' possible to provide a static value for time_coverage_start and'
           ' time_coverage_end (via input file or global overrides)',
           '    time_coverage_relative_end\tOffset (in seconds) that must'
           ' be applied to the values of the time variable to obtain the upper'
           ' bound of the time coverage. This option is required when there is'
           ' more than one time value in the input file or when it is not'
           ' possible to provide a static value for time_coverage_start and'
           ' time_coverage_end (via input file or global overrides)',
           '    climatology\tdirectory where daily climatology files are'
           ' stored if an anomaly is computed',
           )
    out = ('',)
    return ('\n'.join(inp), '\n'.join(out))


def configure_grid_dimensions(output_opts: OutputOptions,
                              granule: Granule,
                              lat: numpy.typing.NDArray,
                              lat_reversed: bool,
                              lon: numpy.typing.NDArray,
                              lon_reversed: bool
                              ) -> typing.Any:
    """"""
    lat_slice = None
    lat_shift = None
    lon_slice = None
    lon_shift = None

    _remove_oob = output_opts.get('remove_out_of_bounds_coordinates', 'false')
    remove_oob = _remove_oob.lower() in ('yes', 'true', '1')
    if remove_oob is False:
        return (lat, lat_slice, lat_shift, lon, lon_slice, lon_shift)

    # Check that bounds are defined
    _west = granule.meta.get('geospatial_lon_min', None)
    west = None
    if _west is not None:
        try:
            west = float(_west)
            if not numpy.isfinite(west):
                west = None
        except ValueError:
            west = None

    _east = granule.meta.get('geospatial_lon_max', None)
    east = None
    if _east is not None:
        try:
            east = float(_east)
            if not numpy.isfinite(east):
                east = None
        except ValueError:
            east = None

    _south = granule.meta.get('geospatial_lat_min', None)
    south = None
    if _south is not None:
        try:
            south = float(_south)
            if not numpy.isfinite(south):
                south = None
        except ValueError:
            south = None

    _north = granule.meta.get('geospatial_lat_max', None)
    north = None
    if _north is not None:
        try:
            north = float(_north)
            if not numpy.isfinite(north):
                north = None
        except ValueError:
            north = None

    has_lat_bounds = (None not in (south, north))
    has_lon_bounds = (None not in (west, east))
    if (has_lat_bounds or has_lon_bounds) is False:
        # No bounds defined
        return (lat, lat_slice, lat_shift, lon, lon_slice, lon_shift)

    # Compute new slice on latitude dimension
    if has_lat_bounds:
        if lat_reversed:
            lat_slice_start = numpy.where(lat <= north)[0][0]
            _lat_slice_end = numpy.where(lat >= south)[0][-1]
        else:
            lat_slice_start = numpy.where(lat >= south)[0][0]
            _lat_slice_end = numpy.where(lat <= north)[0][-1]

        lat_slice_end = None
        if _lat_slice_end < len(lat) - 1:
            lat_slice_end = _lat_slice_end + 1
        lat_slice = slice(lat_slice_start, lat_slice_end, None)

    # Compute new slice on longitude dimension
    if has_lon_bounds:
        if (east is not None) and (west is not None) and (east < west):
            # Data cross the IDL, shift longitudes so that the IDL is not at
            # the edge of the matrix
            lon_180 = numpy.mod(lon + 180.0, 360.0) - 180.0
            lon_shift = -1 * len(numpy.where(0 > lon_180)[0])
            if lon_reversed:
                lon_shift = -1 * lon_shift

            # Update coordinates to use a world view centered on lon 180.0
            lon_sorted = numpy.roll(lon, lon_shift, axis=0)
            lon = numpy.mod(lon_sorted + 0.0, 360.0) - 0.0
            east = numpy.mod(east + 0.0, 360.0) - 0.0
            west = numpy.mod(west + 0.0, 360.0) - 0.0

        if lon_reversed:
            lon_slice_start = numpy.where(lon <= east)[0][0]
            _lon_slice_end = numpy.where(lon >= west)[0][-1]
        else:
            lon_slice_start = numpy.where(lon >= west)[0][0]
            _lon_slice_end = numpy.where(lon <= east)[0][-1]

        lon_slice_end = None
        if _lon_slice_end < len(lon) - 1:
            lon_slice_end = _lon_slice_end + 1
        lon_slice = slice(lon_slice_start, lon_slice_end, None)

    return (lat, lat_slice, lat_shift, lon, lon_slice, lon_shift)


def read_data(input_opts: InputOptions,
              output_opts: OutputOptions
              ) -> typing.Iterator[ReaderResult]:
    """Read input file, extract data and metadata, store them in a Granule
    object then prepare formatting instructions to finalize the conversion to
    IDF format.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule, list)
        A tuple which contains four elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the extracted information
          has been stored

        - a :obj:list of :obj:dict describing the formatting operations that
          the converter must perform before serializing the result in IDF
          format
    """
    _var_ids = input_opts.get('variables', None)
    if _var_ids is None:
        raise MissingVariablesList()
    var_ids = [x.strip() for x in _var_ids.split(',')]

    lon_var_id = input_opts.get('lon_variable', None)
    lat_var_id = input_opts.get('lat_variable', None)
    lon_dim_id = input_opts.get('lon_dimension', lon_var_id)
    lat_dim_id = input_opts.get('lat_dimension', lat_var_id)
    _lon_reversed = input_opts.get('lon_reversed', False)
    if _lon_reversed is not False:
        lon_reversed = _lon_reversed.lower() in ('yes', 'true', '1')
    else:
        lon_reversed = _lon_reversed
    _lat_reversed = input_opts.get('lat_reversed', False)
    if _lat_reversed is not False:
        lat_reversed = _lat_reversed.lower() in ('yes', 'true', '1')
    else:
        lat_reversed = _lat_reversed

    time_var_id = input_opts.get('time_variable', None)

    idf_version = output_opts.get('idf_version', '1.0')
    granule = idf_converter.lib.create_granule(idf_version, DATA_MODEL)
    idf_converter.lib.apply_global_overrides(input_opts, granule)
    idf_converter.lib.apply_var_overrides(input_opts, granule)

    _input_path = input_opts.get('path', None)
    if _input_path is None:
        raise InputPathMissing()
    input_path = os.path.normpath(_input_path)

    # Read variables
    f_handler = netCDF4.Dataset(input_path, 'r')
    idf_converter.lib.extract_global_attributes(f_handler, input_opts, granule)

    # - Time
    dtimes: numpy.typing.NDArray = numpy.array([None, ])
    if time_var_id is not None:
        _t = idf_converter.lib.extract_variable_values(f_handler, time_var_id)
        _t_mask = numpy.ma.getmaskarray(_t).copy()
        if _t_mask.all():
            raise MissingValidTime(time_var_id)

        if _t_mask.any():
            logger.warning('Some time values are masked, this is considered '
                           'bad practice. These masked values will be ignored')

            # Temporarily replace masked values by a valid one ortherwise the
            # netCDF4.num2date call will fail. The mask of the time variable
            # will be applied to the output of num2date later on.
            # At this point in the code we know that there is at least one non-
            # masked value, so we use the first we find as a placeholder.
            masked_t_ind = numpy.where(_t_mask)
            first_valid_t = _t[numpy.where(~_t_mask)[0][0]]
            _t[masked_t_ind] = first_valid_t

        _t_units = f_handler.variables[time_var_id].units
        if 'time_units' in input_opts:
            _t_units = input_opts['time_units']
        if 'time_offset' in input_opts:
            _time_offset = float(input_opts['time_offset'])
            _t = _t + _time_offset
        dtimes = netCDF4.num2date(numpy.ma.getdata(_t), _t_units)
        dtimes = idf_converter.lib.time.as_datetimes_array(dtimes)

        # Apply the same mask as the time value read from the input file to
        # ignore cells that were originally masked and were temporarily
        # replaced by valid values to avoid computation issues
        dtimes = numpy.ma.masked_where(_t_mask, dtimes, copy=False)

    # - Space
    _geoloc_at_pixel_center = input_opts.get('geoloc_at_pixel_center',
                                             'true').lower()
    geoloc_at_pixel_center = _geoloc_at_pixel_center in ('true', 'yes', '1')
    if lat_var_id is not None:
        dim1 = idf_converter.lib.extract_variable_values(f_handler, lat_var_id,
                                                         False)
    elif lat_dim_id is not None:
        dim1_max = None
        dim1_min = None
        if hasattr(f_handler, 'northernmost_latitude'):
            dim1_max = float(f_handler.northernmost_latitude)
        elif hasattr(f_handler, 'geospatial_lat_max'):
            dim1_max = float(f_handler.geospatial_lat_max)
        if hasattr(f_handler, 'southernmost_latitude'):
            dim1_min = float(f_handler.southernmost_latitude)
        elif hasattr(f_handler, 'geospatial_lat_min'):
            dim1_min = float(f_handler.geospatial_lat_min)

        if (dim1_min is None) or (dim1_max is None):
            raise MissingLatitudeDimension()

        dim1_size = f_handler.dimensions[lat_dim_id].size
        dlat = (dim1_max - dim1_min) / dim1_size
        _dim1 = [dim1_min + dlat * i for i in range(0, dim1_size)]
        dim1 = numpy.array(_dim1)
        if geoloc_at_pixel_center is True:
            dim1 = dim1 + 0.5 * dlat
    else:
        raise MissingLatitudeDimension()

    if lon_var_id is not None:
        dim2 = idf_converter.lib.extract_variable_values(f_handler, lon_var_id,
                                                         False)
    elif lon_dim_id is not None:
        dim2_max = None
        dim2_min = None
        if hasattr(f_handler, 'easternmost_longitude'):
            dim2_max = float(f_handler.easternmost_longitude)
        elif hasattr(f_handler, 'geospatial_lon_max'):
            dim2_max = float(f_handler.geospatial_lon_max)
        if hasattr(f_handler, 'westernmost_longitude'):
            dim2_min = float(f_handler.westernmost_longitude)
        elif hasattr(f_handler, 'geospatial_lon_min'):
            dim2_min = float(f_handler.geospatial_lon_min)

        if (dim2_min is None) or (dim2_max is None):
            raise MissingLongitudeDimension()

        dim2_size = f_handler.dimensions[lon_dim_id].size
        dlon = (dim2_max - dim2_min) / dim2_size
        _dim2 = [dim2_min + dlon * i for i in range(0, dim2_size)]
        dim2 = numpy.array(_dim2)
        if geoloc_at_pixel_center is True:
            dim2 = dim2 + 0.5 * dlon
    else:
        raise MissingLongitudeDimension()

    dim1_shape = numpy.shape(dim1)
    while (2 < len(dim1_shape)) and 1 == dim1_shape[0]:
        dim1 = dim1[0]
        dim1_shape = numpy.shape(dim1)

    dim2_shape = numpy.shape(dim2)
    while (2 < len(dim2_shape)) and 1 == dim2_shape[0]:
        dim2 = dim2[0]
        dim2_shape = numpy.shape(dim2)

    # The geospatial_lon|lat_min|max attributes have been extracted and set on
    # the granule when calling extract_global_attributes but some files use
    # east|westernmost_longitude and south|northernmost_latitude instead and
    # these attributes are not extracted automatically (not part of IDF v1
    # specs).
    # Global overrides have already been applied, so use these attributes only
    # if the geospatial_lon|lat_min|max attributes are not already set on the
    # granule.
    if 'geospatial_lon_max' not in granule.meta.keys():
        if hasattr(f_handler, 'easternmost_longitude'):
            value = f_handler.easternmost_longitude
            granule.meta['geospatial_lon_max'] = value
    if 'geospatial_lon_min' not in granule.meta.keys():
        if hasattr(f_handler, 'westernmost_longitude'):
            value = f_handler.westernmost_longitude
            granule.meta['geospatial_lon_min'] = value
    if 'geospatial_lat_max' not in granule.meta.keys():
        if hasattr(f_handler, 'northernmost_latitude'):
            value = f_handler.northernmost_latitude
            granule.meta['geospatial_lat_max'] = value
    if 'geospatial_lat_min' not in granule.meta.keys():
        if hasattr(f_handler, 'southernmost_latitude'):
            value = f_handler.southernmost_latitude
            granule.meta['geospatial_lat_min'] = value

    dims_config = configure_grid_dimensions(output_opts, granule,
                                            dim1, lat_reversed,
                                            dim2, lon_reversed)
    dim1, dim1_slice, dim1_shift, dim2, dim2_slice, dim2_shift = dims_config

    if dim1_slice is not None:
        dim1 = dim1[dim1_slice]
    dim1_size = numpy.shape(dim1)[0]

    if dim2_slice is not None:
        dim2 = dim2[dim2_slice]
    dim2_size = numpy.shape(dim2)[0]

    dim1_info = (lat_dim_id, lat_reversed, dim1_shift, dim1_slice)
    dim2_info = (lon_dim_id, lon_reversed, dim2_shift, dim2_slice)

    """
    # Wrap data in [-180.0, 180[ instead of [0, 360[
    # This is not a requirement of the IDF specifications but the current
    # implementation of streamlines rendering in SEAScope will not work
    # otherwise
    wrap_lon_180 = input_opts.get('wrap_lon_180', 'true').lower()
    if wrap_lon_180 in ('true', 'yes', '1'):
        dim2 = numpy.mod(dim2 + 180.0, 360.0) - 180.0
        negative_lon_ind = numpy.where(dim2 < 0)[0]
        if negative_lon_ind.any():
            lon_shift = -negative_lon_ind[0]
        dim2 = numpy.roll(dim2, lon_shift)
    """

    # - Depth
    depth_info = idf_converter.lib.generic_grid.get_depth_info(input_opts,
                                                               granule,
                                                               f_handler)

    granule.vars['lat'] = {'array': dim1,
                           'units': 'degrees_north',
                           'datatype': dim1.dtype,
                           'options': {}}
    granule.vars['lon'] = {'array': dim2,
                           'units': 'degrees_east',
                           'datatype': dim2.dtype,
                           'options': {}}
    granule.dims['lat'] = dim1_size
    granule.dims['lon'] = dim2_size

    dlon = dim2[1] - dim2[0]
    dlat = dim1[1] - dim1[0]
    spatial_resolution = min([abs(dlat), abs(dlon)]) * 111000.

    granule.meta['idf_subsampling_factor'] = 0
    granule.meta['idf_spatial_resolution'] = spatial_resolution
    granule.meta['idf_spatial_resolution_units'] = 'm'

    granule_file = os.path.basename(input_path)
    granule_name, _ = os.path.splitext(granule_file)
    granule.meta['idf_granule_id'] = granule_name

    if 'indicative_times' in input_opts:
        dtimes = input_opts['indicative_times']
    generic_processor = idf_converter.lib.generic_grid.single_granule
    if 1 < len(dtimes):
        generic_processor = idf_converter.lib.generic_grid.multi_granule
    for extracted_data in generic_processor(granule, f_handler, var_ids,
                                            depth_info, dim1_info, dim2_info,
                                            input_opts, output_opts,
                                            time_var_id, dtimes):
        yield extracted_data

    f_handler.close()



# vim: ts=4:sts=4:sw=4
#
# @date 2019-09-27
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import os
import numpy
import numpy.typing
import typing
import netCDF4
import logging
import idf_converter.lib
from idf_converter.lib.types import InputOptions, OutputOptions, Granule
from idf_converter.lib.types import ReaderResult, AtmCorrData, NumValuesDict
from idf_converter.lib.types import TransformsList

logger = logging.getLogger(__name__)

DATA_MODEL = 'SWATH'
MASK_BITS = {'saturated@Oa21': 0,
             'saturated@Oa20': 1,
             'saturated@Oa19': 2,
             'saturated@Oa18': 3,
             'saturated@Oa17': 4,
             'saturated@Oa16': 5,
             'saturated@Oa15': 6,
             'saturated@Oa14': 7,
             'saturated@Oa13': 8,
             'saturated@Oa12': 9,
             'saturated@Oa11': 10,
             'saturated@Oa10': 11,
             'saturated@Oa09': 12,
             'saturated@Oa08': 13,
             'saturated@Oa07': 14,
             'saturated@Oa06': 15,
             'saturated@Oa05': 16,
             'saturated@Oa04': 17,
             'saturated@Oa03': 18,
             'saturated@Oa02': 19,
             'saturated@Oa01': 20,
             'dubious': 21,
             'sun-glint_risk': 22,
             'duplicated': 23,
             'cosmetic': 24,
             'invalid': 25,
             'straylight_risk': 26,
             'bright': 27,
             'tidal_region': 28,
             'fresh_inland_water': 29,
             'coastline': 30,
             'land': 31}

# S-2 : 665, 560, 490
# red : 650 nm
# green : 510 nm
# blue : 475 nm
# https://sentinel.esa.int/web/sentinel/user-guides/sentinel-3-olci/overview/heritage
# https://sentinel.esa.int/web/sentinel/technical-guides/sentinel-3-olci/level-1/fr-or-rr-toa-radiances
# Oa1   400 	15
# Oa2 	412.5 	10
# Oa3 	442.5 	10
# Oa4 	490 	10   BLUE
# Oa5 	510 	10
# Oa6 	560 	10   GREEN
# Oa7 	620 	10
# Oa8 	665 	10   RED
# Oa9 	673.75 	7.5
# Oa10 	681.25 	7.5
# Oa11 	708.75 	10
# Oa12 	753.75 	7.5
# Oa13 	761.25 	2.5
# Oa14 	764.375 3.75
# Oa15 	767.5 	2.5
# Oa16 	778.75 	15
# Oa17 	865 	20 SUN GLITTER
# Oa18 	885 	10
# Oa19 	900 	10
# Oa20 	940 	20
# Oa21 	1 020 	40


class InputPathMissing(IOError):
    """"""
    pass


class GeoCoordinatesNotFound(IOError):
    """"""
    pass


class TimeCoordinatesNotFound(IOError):
    """"""
    pass


class QualityFlagsNotFound(IOError):
    """"""
    pass


class RadianceFileNotFound(IOError):
    """"""
    pass


class TieGeometriesNotFound(IOError):
    """"""
    pass


class InstrumentDataNotFound(IOError):
    """"""
    pass


class LookUpTableNotFound(IOError):
    """"""
    pass


class IncompatibleGeometryDetected(ValueError):
    """Error raised when the alongtrack dimension in the tie geometry file
    differs from the alongtrack dimension of the geo coordinates file."""
    pass


def get_brightness_mask(quality_path: str
                        ) -> numpy.typing.NDArray:
    """"""
    off_invalid = numpy.uint32(1 << MASK_BITS['invalid'])
    off_fresh = numpy.uint32(1 << MASK_BITS['fresh_inland_water'])
    off_land = numpy.uint32(1 << MASK_BITS['land'])

    # Read quality flags
    quality_handler = netCDF4.Dataset(quality_path, 'r')
    _quality_flags = idf_converter.lib.extract_variable_values(quality_handler,
                                                               'quality_flags')
    quality_flags = numpy.ma.array(_quality_flags)
    quality_handler.close()

    mask: numpy.typing.NDArray
    mask = ((numpy.bitwise_and(quality_flags, off_invalid) > 0) |
            ((numpy.bitwise_and(quality_flags, off_land) > 0) &
             (numpy.bitwise_and(quality_flags, off_fresh) == 0)))

    return mask


def help() -> typing.Tuple[str, str]:
    """Describe options supported by this reader.

    Returns
    -------
    str
        Description of the supported options
    """
    inp = ('    path\tPath of the input file',
           '    atmospheric_correction_LUT_path\tPath of the Rayleigh'
           ' scattering LUT (CNES)')
    out = ('',)
    return ('\n'.join(inp), '\t'.join(out))


def get_contrast_common_mask(quality_path: str,
                             channels: typing.Iterable[str]
                             ) -> numpy.typing.NDArray:
    """"""
    off_flags = numpy.uint32(0)
    off_flags = off_flags + numpy.uint32(1 << MASK_BITS['dubious'])
    off_flags = off_flags + numpy.uint32(1 << MASK_BITS['invalid'])
    off_flags = off_flags + numpy.uint32(1 << MASK_BITS['straylight_risk'])
    off_flags = off_flags + numpy.uint32(1 << MASK_BITS['bright'])
    off_flags = off_flags + numpy.uint32(1 << MASK_BITS['tidal_region'])
    off_flags = off_flags + numpy.uint32(1 << MASK_BITS['coastline'])
    off_flags = off_flags + numpy.uint32(1 << MASK_BITS['land'])

    # We keep inland waters
    # off_flags += numpy.uint32(1 << bits['fresh_inland_water'])

    # Read quality flags
    quality_handler = netCDF4.Dataset(quality_path, 'r')
    _quality_flags = idf_converter.lib.extract_variable_values(quality_handler,
                                                               'quality_flags')
    quality_flags = numpy.ma.array(_quality_flags)
    quality_handler.close()

    contrast_mask = numpy.zeros(quality_flags.shape, dtype='bool')
    contrast_mask = (contrast_mask |
                     (numpy.bitwise_and(quality_flags, off_flags) > 0))
    return contrast_mask


def get_contrast_bands_mask(quality_path: str,
                            channels: typing.Iterable[str]) -> NumValuesDict:
    """"""
    # Read quality flags
    quality_handler = netCDF4.Dataset(quality_path, 'r')
    _quality_flags = idf_converter.lib.extract_variable_values(quality_handler,
                                                               'quality_flags')
    quality_flags = numpy.ma.array(_quality_flags)
    quality_handler.close()

    # Filter out values where any of the bands is flagged as saturated
    masks = {}
    for band_name in channels:
        bit_name = 'saturated@{}'.format(band_name[0:4])
        off_flags = numpy.uint32(0)
        off_flags = off_flags + numpy.uint32(1 << MASK_BITS[bit_name])

        _mask = numpy.zeros(quality_flags.shape, dtype='bool')
        _mask = (_mask | (numpy.bitwise_and(quality_flags, off_flags) > 0))
        masks[band_name] = _mask
    return masks


def get_atm_corr_data(input_dir: str, lut_path: str,
                      nrow: int, ncell: int) -> AtmCorrData:
    """"""
    if not os.path.exists(lut_path):
        raise LookUpTableNotFound(lut_path)

    instrument_path = os.path.join(input_dir, 'instrument_data.nc')
    if not os.path.exists(instrument_path):
        raise InstrumentDataNotFound(instrument_path)

    tie_geometries_path = os.path.join(input_dir, 'tie_geometries.nc')
    if not os.path.exists(tie_geometries_path):
        raise TieGeometriesNotFound(tie_geometries_path)

    bands_count = 21  # Oa01 -> Oa21

    # Read Look-Up Table
    sza_lut = []
    oza_lut = []
    delta_lut = []
    rho_atm = []
    with open(lut_path, 'r') as f:
        for line in f.readlines():
            if line.startswith('#'):
                continue
            split_line = line.split(' ')
            sza_lut.append(split_line[0])
            oza_lut.append(split_line[1])
            delta_lut.append(split_line[2])
            rho_line = []
            for icolband in range(3, 3 + bands_count):
                rho_line.append(float(split_line[icolband]))
            rho_atm.append(rho_line)

    # Bands numbering starts at 1 and are name "Oa__" where __ is the band
    # number (2 digits)
    bands_index = {f'Oa{x + 1:02}_radiance': x for x in range(0, bands_count)}

    # Read angles from tie_geometries.nc
    f_handler = netCDF4.Dataset(tie_geometries_path, 'r')
    sza = idf_converter.lib.extract_variable_values(f_handler, 'SZA')
    saa = idf_converter.lib.extract_variable_values(f_handler, 'SAA')
    oaa = idf_converter.lib.extract_variable_values(f_handler, 'OAA')
    oza = idf_converter.lib.extract_variable_values(f_handler, 'OZA')
    ac_subsampling = f_handler.ac_subsampling_factor
    f_handler.close()

    nrow_geom, ncell_geom = numpy.shape(sza)
    if nrow_geom != nrow:
        raise IncompatibleGeometryDetected()

    # Read Extra-terrestrial solar irradiance
    solar_flux = {}
    f_handler = netCDF4.Dataset(instrument_path, 'r')
    _solar_flux = idf_converter.lib.extract_variable_values(f_handler,
                                                            'solar_flux')
    for band_index in range(0, bands_count):
        band_name = f'Oa{band_index + 1:02}_radiance'
        solar_flux[band_name] = numpy.ma.array(_solar_flux[band_index])
    f_handler.close()

    atm_corr_data = {'SZA_LUT': sza_lut,
                     'OZA_LUT': oza_lut,
                     'DELTA_LUT': delta_lut,
                     'RHO_LUT': rho_atm,
                     'bands_count': bands_count,
                     'bands_index': bands_index,
                     'nrow': nrow,
                     'ncell': ncell,
                     'SZA': numpy.ma.array(sza),
                     'SAA': numpy.ma.array(saa),
                     'OAA': numpy.ma.array(oaa),
                     'OZA': numpy.ma.array(oza),
                     'solar_flux': solar_flux,
                     'al_subsampled': False,
                     'ac_subsampling': ac_subsampling}

    return atm_corr_data


def get_dynamic_bands_mask(granule: Granule,
                           targets: typing.List[str]) -> NumValuesDict:
    """"""
    masks = {}
    for var_id in targets:
        var_data = granule.vars[var_id]['array']

        # Mask null and negative values: they are inferior or equal to
        # atmospheric correction and should probably have been flagged as
        # clouds.
        # They should therefore not be used to computed contrast for ocean
        masks[var_id] = (numpy.ma.getmaskarray(var_data) |
                         (var_data <= 0))
    return masks


def read_data(input_opts: InputOptions,
              output_opts: OutputOptions
              ) -> typing.Iterator[ReaderResult]:
    """Read input file, extract data and metadata, store them in a Granule
    object then prepare formatting instructions to finalize the conversion to
    IDF format.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule, list)
        A tuple which contains four elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the extracted information
          has been stored

        - a :obj:list of :obj:dict describing the formatting operations that
          the converter must perform before serializing the result in IDF
          format
    """
    variables = {'Oa04_radiance': {'name': 'Oa04',
                                   'valid_min': 13.966,
                                   'valid_max': 95.000,
                                   'options': {}},
                 'Oa06_radiance': {'name': 'Oa06',
                                   'valid_min': 8.213,
                                   'valid_max': 95.000,
                                   'options': {}},
                 'Oa09_radiance': {'name': 'Oa09',
                                   'valid_min': 3.346,
                                   'valid_max': 95.000,
                                   'options': {}},
                 'Oa17_radiance': {'name': 'Oa17',
                                   'valid_min': 3.000,
                                   'valid_max': 150.000,
                                   'options': {}},
                 'Oa17_radiance_bc': {'name': 'Oa17_bc',
                                      'long_name': ('Brightness contrast for '
                                                    'OLCI acquisition band '
                                                    'Oa17'),
                                      'units': '',
                                      'valid_min': numpy.exp(-1.2),
                                      'valid_max': numpy.exp(1.5),
                                      'options': {}}}

    visible_vars = ('Oa04_radiance', 'Oa06_radiance', 'Oa09_radiance')
    infrared_vars = ('Oa17_radiance',)
    radiance_vars = ('Oa04_radiance', 'Oa06_radiance', 'Oa09_radiance',
                     'Oa17_radiance')
    channels = list(radiance_vars)

    idf_version = output_opts.get('idf_version', '1.0')
    granule = idf_converter.lib.create_granule(idf_version, DATA_MODEL)
    granule.vars = variables

    _input_dir = input_opts.get('path', None)
    if _input_dir is None:
        raise InputPathMissing()
    input_dir = os.path.normpath(_input_dir)

    # Check existance of required files
    geo_path = os.path.join(input_dir, 'geo_coordinates.nc')
    if not os.path.exists(geo_path):
        raise GeoCoordinatesNotFound(geo_path)

    time_path = os.path.join(input_dir, 'time_coordinates.nc')
    if not os.path.exists(time_path):
        raise TimeCoordinatesNotFound(time_path)

    quality_path = os.path.join(input_dir, 'qualityFlags.nc')
    if not os.path.exists(quality_path):
        raise QualityFlagsNotFound(quality_path)

    atm_corr_lut_path = input_opts.get('atmospheric_correction_LUT_path', None)
    atm_correction_requested = atm_corr_lut_path is not None

    compute_nir_nb = True

    # Read variables
    for var_id in radiance_vars:
        file_name = f'{var_id}.nc'
        radiance_path = os.path.join(input_dir, file_name)
        if not os.path.exists(radiance_path):
            raise RadianceFileNotFound(radiance_path)

        f_handler = netCDF4.Dataset(radiance_path, 'r')
        idf_converter.lib.extract_variable_attributes(f_handler, var_id,
                                                      granule)
        band = idf_converter.lib.extract_variable_values(f_handler, var_id)
        granule.vars[var_id]['array'] = numpy.ma.array(band)
        granule.vars[var_id]['datatype'] = band.dtype
        f_handler.close()

    # Extract geo coordinates information
    geo_handler = netCDF4.Dataset(geo_path, 'r')
    nrow = geo_handler.dimensions['rows'].size
    ncell = geo_handler.dimensions['columns'].size
    lon = idf_converter.lib.extract_variable_values(geo_handler, 'longitude')
    lon = numpy.ma.array(lon)
    lat = idf_converter.lib.extract_variable_values(geo_handler, 'latitude')
    lat = numpy.ma.array(lat)
    geo_handler.close()

    granule.dims['row'] = nrow
    granule.dims['cell'] = ncell

    # Handle longitude continuity
    dlon = lon[1:, :] - lon[:-1, :]
    if 180.0 <= numpy.max(numpy.abs(dlon)):
        lon[lon < 0.0] = lon[lon < 0.0] + 360.0

    # Extract time coordinates information
    time_handler = netCDF4.Dataset(time_path, 'r')
    idf_converter.lib.extract_global_attributes(time_handler, input_opts,
                                                granule)
    _time = idf_converter.lib.extract_variable_values(time_handler,
                                                      'time_stamp')
    start_timestamp = _time[0]
    end_timestamp = _time[-1]
    timestamp_units = time_handler.variables['time_stamp'].units
    time_handler.close()

    granule.vars['lat'] = {'array': lat,
                           'units': 'degrees_north',
                           'datatype': lat.dtype,
                           'options': {}}
    granule.vars['lon'] = {'array': lon,
                           'units': 'degrees_east',
                           'datatype': lon.dtype,
                           'options': {}}

    # Format time information
    start_dt = netCDF4.num2date(start_timestamp, timestamp_units)
    stop_dt = netCDF4.num2date(end_timestamp, timestamp_units)

    granule_name = os.path.basename(input_dir)

    granule.meta['idf_granule_id'] = granule_name
    granule.meta['idf_subsampling_factor'] = 0
    granule.meta['idf_spatial_resolution'] = 300
    granule.meta['idf_spatial_resolution_units'] = 'm'
    granule.meta['time_coverage_start'] = start_dt
    granule.meta['time_coverage_end'] = stop_dt
    granule.meta['institution'] = 'ESA'
    granule.meta['platform'] = 'Sentinel-3'
    granule.meta['platform_type'] = 'leo satellite'
    granule.meta['sensor'] = 'OLCI'
    granule.meta['sensor_type'] = 'medium-resolution imaging spectrometer'

    # Add transforms
    transforms: TransformsList = []

    # APPLY ATMOSPHERIC CORRECTION
    if atm_correction_requested:
        atm_corr_data = get_atm_corr_data(input_dir, atm_corr_lut_path, nrow,
                                          ncell)

        # Add brightness contrast
        if compute_nir_nb:
            bc_mask = get_brightness_mask(quality_path)

            transforms.append(('s3_nir_bc', {'targets': infrared_vars,
                                             'atm_corr_data': atm_corr_data}))

            bc_channels = [f'{channel}_bc' for channel in infrared_vars]
            transforms.append(('static_common_mask', {'targets': bc_channels,
                                                      'mask': bc_mask}))
            channels.extend(bc_channels)

        transforms.append(('s3_cnes_rayleigh',
                           {'targets': radiance_vars,
                            'atm_corr_data': atm_corr_data}))

    # Compute extrema depending on valid data proportion
    # Use default values if there is not enough valid data
    # Use percentiles otherwise
    ignored_common_mask = get_contrast_common_mask(quality_path,
                                                   radiance_vars)
    ignored_bands_mask = get_contrast_bands_mask(quality_path,
                                                 radiance_vars)
    transforms.append(('contrast_from_pct',
                       {'targets': radiance_vars,
                        'valid_threshold': 0.001,  # 0.1%
                        'common_mask': ignored_common_mask,
                        'bands_mask': ignored_bands_mask,
                        'dynamic_bands_mask': get_dynamic_bands_mask,
                        'min_percentile': 0.5,
                        'max_percentile': 99.99}))

    # Share extrema among visible channels
    transforms.append(('share_extrema', {'targets': visible_vars}))

    # Artificially increase the range on a linear scale, thus allowing
    # users to desaturate the rendering of this channel in SEAScope
    # _max_linear = _min_linear + 5 * (_max_linear - _min_linear)
    transforms.append(('stretch_extent', {'targets': radiance_vars,
                                          'stretch_factor': 5.0,
                                          'change_vmin': False,
                                          'change_vmax': True}))

    # APPLY LOGARITHMIC SCALE
    # Pixels with a radiance equal or inferior to atmospheric correction
    # will be clipped to the minimal value as logscale cannot be applied to
    # negative values.
    transforms.append(('logscale', {'targets': channels,
                                    'base': 0}))

    # TODO: it was necessary for Syntool but is it relevant for SEAScope?
    # transforms.append(('share_masks', {'targets': channels}))

    output_opts['__export'] = channels

    yield input_opts, output_opts, granule, transforms



# vim: ts=4:sts=4:sw=4
#
# @date 2019-09-30
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import os
import numpy
import numpy.typing
import typing
import netCDF4
import logging
import datetime
import idf_converter.lib
from idf_converter.lib import EarlyExit
from idf_converter.lib.types import InputOptions, OutputOptions, Granule
from idf_converter.lib.types import ReaderResult, AtmCorrData, NumValuesDict
from idf_converter.lib.types import Extrema, TransformsList

logger = logging.getLogger(__name__)

DATA_MODEL = 'SWATH'
MAX_SUNGLINT = 150.000


class InputPathMissing(IOError):
    """Error rasied when the "path" input option has not been specified by the
    user."""
    pass


class GeoCoordinatesNotFound(IOError):
    """Error raised when the path for the geographical coordinates file does
    not exist."""
    def __init__(self, path: str) -> None:
        """"""
        self.path = path


class RadianceFileNotFound(IOError):
    """Error raised when the path for the file which contains radiance values
    does not exist."""
    def __init__(self, path: str) -> None:
        """"""
        self.path = path


class LookUpTableNotFound(IOError):
    """Error raised when the path for the lookup table file does not exist."""
    def __init__(self, path: str) -> None:
        """"""
        self.path = path


class TieGeometriesNotFound(IOError):
    """Error raised when the path for the tie geometries file does not exist.
    """
    def __init__(self, path: str) -> None:
        """"""
        self.path = path


class UnsupportedChannels(Exception):
    def __init__(self, channels: typing.Iterable[str],
                 *args: typing.Iterable[typing.Any]) -> None:
        _chans = ', '.join(channels)
        msg = 'The following channels are not supported: {}'.format(_chans)
        super(UnsupportedChannels, self).__init__(msg, *args)


class ChannelsMissing(Exception):
    pass


class MissingRequiredFiles(IOError):
    pass


class OnlyMaskedValues(Exception):
    pass


class OnlyNightData(Exception):
    pass


class CorruptedAtmosphericCorrection(Exception):
    def __init__(self) -> None:
        msg = ('Infinite or nan value found in the atmospheric correction, '
               'please check that the LUT has valid values for the angles '
               'contained in the geometrie_tn.nc file')
        super(CorruptedAtmosphericCorrection, self).__init__(msg)


def help() -> typing.Tuple[str, str]:
    """Describe options supported by this reader.

    Returns
    -------
    str
        Description of the supported options
    """
    inp = ('    path\tPath of the input file',
           '    atmospheric_correction_LUT_path\tPath of the Rayleigh'
           ' scattering LUT (CNES)')
    out = ('',)
    return ('\n'.join(inp), '\n'.join(out))


def fix_extrema(var_dict: typing.Dict[str, typing.Any],
                default_min: float,
                default_max: float) -> Extrema:
    """"""
    _min = var_dict['valid_min']
    _max = var_dict['valid_max']
    if 100.000 < _min:
        _min = default_min
        _max = default_max
    return _min, _max


def build_masks(input_dir: str
                ) -> typing.Tuple[numpy.typing.NDArray, numpy.typing.NDArray]:
    """"""

    bits = {'coastline': 0,
            'ocean': 1,
            'tidal': 2,
            'land': 3,
            'inland_water': 4,
            'unfilled': 5,
            'spare': 6,
            'spare_': 7,
            'cosmetic': 8,
            'duplicate': 9,
            'day': 10,
            'twilight': 11,
            'sun_glint': 12,
            'snow': 13,
            'summary_cloud': 14,
            'summary_pointing': 15}

    # Invalidate if these flags are all off
    on_flags = numpy.ushort(0)
    on_flags = on_flags + numpy.ushort(1 << bits['day'])

    # Invalidate if any of these flags is on
    off_flags = numpy.ushort(0)
    off_flags = off_flags + numpy.ushort(1 << bits['coastline'])
    off_flags = off_flags + numpy.ushort(1 << bits['land'])
    off_flags = off_flags + numpy.ushort(1 << bits['snow'])

    # Read quality flags
    quality_path = os.path.join(input_dir, 'flags_an.nc')
    quality_handler = netCDF4.Dataset(quality_path, 'r')
    _quality_flags = idf_converter.lib.extract_variable_values(quality_handler,
                                                               'confidence_an')
    quality_flags = numpy.ma.array(_quality_flags)
    quality_handler.close()

    # Intialize mask to compute histograms on selected valid values
    contrast_mask = numpy.zeros(quality_flags.shape, dtype='bool')

    # Initialize mask to remove invalid data
    data_mask = numpy.zeros(quality_flags.shape, dtype='bool')

    # Mask data under sun_glint too
    off_flags = off_flags + numpy.uint32(1 << bits['sun_glint'])

    contrast_mask = (contrast_mask |
                     (numpy.bitwise_and(quality_flags, on_flags) <= 0) |
                     (numpy.bitwise_and(quality_flags, off_flags) > 0))

    # Granules containing only night/twilight data must be ignored.
    # Twilight data are only useful to improve the contrast when there is
    # at least some daily data.
    on_flags = numpy.ushort(0)
    on_flags |= numpy.ushort(1 << bits['day'])
    data_mask = (data_mask |
                 (numpy.bitwise_and(quality_flags,  on_flags) <= 0))
    if data_mask.all():
        raise OnlyNightData()

    return contrast_mask, data_mask


def get_bands_contrast_mask(granule: Granule,
                            targets: typing.List[str]) -> NumValuesDict:
    """"""
    masks = {}
    for var_id in targets:
        var_data = granule.vars[var_id]['array']

        # Ignore sunglint when computing the contrast for RGB channels
        sunglint_mask = (var_data >= MAX_SUNGLINT)
        negative_mask = (var_data <= 0)
        data_mask = (numpy.ma.getmaskarray(var_data))
        masks[var_id] = (data_mask | negative_mask | sunglint_mask)

    return masks


def get_atm_corr_data(input_dir: str, lut_path: str,
                      nrow: int, ncell: int) -> AtmCorrData:
    """"""
    view = 'n'  # Nadir view

    if not os.path.exists(lut_path):
        raise LookUpTableNotFound(lut_path)

    tie_geometries_path = os.path.join(input_dir, f'geometry_t{view}.nc')
    if not os.path.exists(tie_geometries_path):
        raise TieGeometriesNotFound(tie_geometries_path)

    bands_count = 9  # S1 -> S9

    # Read Look-Up Table
    sza_lut = []
    oza_lut = []
    delta_lut = []
    rho_atm = []
    with open(lut_path, 'r') as f:
        for line in f.readlines():
            if line.startswith('#'):
                continue
            split_line = line.split(' ')
            sza_lut.append(split_line[0])
            oza_lut.append(split_line[1])
            delta_lut.append(split_line[2])
            rho_line = []
            for icolband in range(3, 3 + bands_count):
                rho_line.append(float(split_line[icolband]))
            rho_atm.append(rho_line)

    # Bands numbering starts at 1 and are name "S_" where _ is the band
    # number
    bands_index = {f'S{x + 1}_radiance_an': x for x in range(0, bands_count)}

    # Read angles from geometry_t*.nc
    f_handler = netCDF4.Dataset(tie_geometries_path, 'r')
    sza = idf_converter.lib.extract_variable_values(f_handler,
                                                    f'solar_zenith_t{view}')
    saa = idf_converter.lib.extract_variable_values(f_handler,
                                                    f'solar_azimuth_t{view}')
    oaa = idf_converter.lib.extract_variable_values(f_handler,
                                                    f'sat_azimuth_t{view}')
    oza = idf_converter.lib.extract_variable_values(f_handler,
                                                    f'sat_zenith_t{view}')
    f_handler.close()

    # if sza is more than 90 take values from 80 - 85 range
    sza_nan_ind = numpy.where(numpy.isnan(sza))
    sza[sza_nan_ind] = -1  # temporary, to avoid warnings with "sza > 84"
    sza[sza > 84] = 84
    sza[sza_nan_ind] = numpy.nan

    # Replace nan values by the last valid one
    ifirst = 19
    ilast = ifirst + 1
    sza[:, -ifirst:] = numpy.repeat(sza[:, -ilast],
                                    ifirst).reshape((sza.shape[0], ifirst))
    saa[:, -ifirst:] = numpy.repeat(saa[:, -ilast],
                                    ifirst).reshape((saa.shape[0], ifirst))
    oza[:, -ifirst:] = numpy.repeat(oza[:, -ilast],
                                    ifirst).reshape((oza.shape[0], ifirst))
    oaa[:, -ifirst:] = numpy.repeat(oaa[:, -ilast],
                                    ifirst).reshape((oaa.shape[0], ifirst))

    # Read Extra-terrestrial solar irradiance
    solar_flux = {}
    _get_values = idf_converter.lib.extract_variable_values
    for band_index in range(0, bands_count):
        band_number = band_index + 1
        band_name = f'S{band_number}_radiance_an'
        if band_index > 5:  # band_index = 6 <=> band_name = S7
            continue
        file_name = f'S{band_number}_quality_an.nc'
        file_path = os.path.join(input_dir, file_name)
        f_handler = netCDF4.Dataset(file_path, 'r')
        var_name = f'S{band_number}_solar_irradiance_an'
        solar_flux[band_name] = _get_values(f_handler, var_name)
        f_handler.close()

    nrow_geom, ncell_geom = numpy.shape(sza)
    if abs(nrow_geom - nrow / 2.0) > 1:
        logger.warning('Discrepency between channel and geometry dimensions')

    atm_corr_data = {'SZA_LUT': sza_lut,
                     'OZA_LUT': oza_lut,
                     'DELTA_LUT': delta_lut,
                     'RHO_LUT': rho_atm,
                     'bands_count': bands_count,
                     'bands_index': bands_index,
                     'nrow': nrow,
                     'ncell': ncell,
                     'SZA': numpy.ma.array(sza),
                     'SAA': numpy.ma.array(saa),
                     'OAA': numpy.ma.array(oaa),
                     'OZA': numpy.ma.array(oza),
                     'solar_flux': solar_flux,
                     'al_subsampled': True,
                     'ac_subsampling': int(ncell / ncell_geom)}

    return atm_corr_data


def read_data(input_opts: InputOptions,
              output_opts: OutputOptions
              ) -> typing.Iterator[ReaderResult]:
    """Read input file, extract data and metadata, store them in a Granule
    object then prepare formatting instructions to finalize the conversion to
    IDF format.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule, list)
        A tuple which contains four elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the extracted information
          has been stored

        - a :obj:list of :obj:dict describing the formatting operations that
          the converter must perform before serializing the result in IDF
          format
    """
    variables = {'S1_radiance_an': {'name': 'S1',
                                    'valid_min': 14.000,
                                    'valid_max': 150.000,
                                    'options': {}},
                 'S2_radiance_an': {'name': 'S2',
                                    'valid_min': 8.000,
                                    'valid_max': 150.000,
                                    'options': {}},
                 'S3_radiance_an': {'name': 'S3',
                                    'valid_min': 4.000,
                                    'valid_max': 150.000,
                                    'options': {}}}

    channels = list(variables.keys())

    idf_version = output_opts.get('idf_version', '1.0')
    granule = idf_converter.lib.create_granule(idf_version, DATA_MODEL)
    granule.vars = variables

    _input_dir = input_opts.get('path', None)
    input_dir = os.path.normpath(_input_dir)
    if input_dir is None:
        raise InputPathMissing()

    view = 'n'  # Process nadir view
    sltype = 'a'

    # Check existance of required files
    geo_path = os.path.join(input_dir, f'geodetic_{sltype}{view}.nc')
    if not os.path.exists(geo_path):
        raise GeoCoordinatesNotFound(geo_path)

    atm_corr_lut_path = input_opts.get('atmospheric_correction_LUT_path', None)
    atm_correction_requested = atm_corr_lut_path is not None

    # Read variables
    for var_id in channels:
        file_name = f'{var_id}.nc'
        rad_path = os.path.join(input_dir, file_name)
        if not os.path.exists(rad_path):
            raise RadianceFileNotFound(rad_path)

        f_handler = netCDF4.Dataset(rad_path, 'r')
        idf_converter.lib.extract_variable_attributes(f_handler, var_id,
                                                      granule)
        band = idf_converter.lib.extract_variable_values(f_handler, var_id)
        granule.vars[var_id]['array'] = numpy.ma.array(band)
        f_handler.close()

        band_mask = numpy.ma.getmaskarray(granule.vars[var_id]['array'])
        if numpy.all(band_mask):
            logger.warning(f'All values masked for channel {var_id}')

    # Extract geo coordinates information
    lat_varname = 'latitude_{}{}'.format(sltype, view)
    lon_varname = 'longitude_{}{}'.format(sltype, view)
    geo_handler = netCDF4.Dataset(geo_path, 'r')
    idf_converter.lib.extract_global_attributes(geo_handler, input_opts,
                                                granule)
    nrow = geo_handler.dimensions['rows'].size
    ncell = geo_handler.dimensions['columns'].size
    lon = idf_converter.lib.extract_variable_values(geo_handler, lon_varname)
    lon = numpy.ma.array(lon)
    lat = idf_converter.lib.extract_variable_values(geo_handler, lat_varname)
    lat = numpy.ma.array(lat)
    start_time_str = geo_handler.start_time
    stop_time_str = geo_handler.stop_time
    geo_handler.close()

    granule.dims['row'] = nrow
    granule.dims['cell'] = ncell

    # Handle longitude continuity
    dlon = lon[1:, :] - lon[:-1, :]
    if 180.0 <= numpy.max(numpy.abs(dlon)):
        lon0 = lon[0, 0] + 180.0
        lon[:, :] = numpy.mod(lon[:, :] - lon0, 360.0) + lon0

    granule.vars['lat'] = {'array': lat,
                           'units': 'degrees_north',
                           'datatype': lat.dtype,
                           'options': {}}
    granule.vars['lon'] = {'array': lon,
                           'units': 'degrees_east',
                           'datatype': lon.dtype,
                           'options': {}}

    # Format time information
    time_fmt = '%Y-%m-%dT%H:%M:%S.%fZ'
    start_dt = datetime.datetime.strptime(start_time_str, time_fmt)
    stop_dt = datetime.datetime.strptime(stop_time_str, time_fmt)

    granule_name = os.path.basename(input_dir)

    granule.meta['idf_granule_id'] = granule_name
    granule.meta['idf_subsampling_factor'] = 0
    granule.meta['idf_spatial_resolution'] = 500
    granule.meta['idf_spatial_resolution_units'] = 'm'
    granule.meta['time_coverage_start'] = start_dt
    granule.meta['time_coverage_end'] = stop_dt
    granule.meta['institution'] = 'ESA'
    granule.meta['platform'] = 'Sentinel-3'
    granule.meta['platform_type'] = 'leo satellite'
    granule.meta['sensor'] = 'SLSTR'
    granule.meta['sensor_type'] = 'dual scan temperature radiometer'

    # Add transforms
    transforms: TransformsList = []

    # APPLY ATMOSPHERIC CORRECTION
    if atm_correction_requested:
        atm_corr_data = get_atm_corr_data(input_dir, atm_corr_lut_path, nrow,
                                          ncell)
        transforms.append(('s3_cnes_rayleigh',
                           {'targets': channels,
                            'atm_corr_data': atm_corr_data}))

    # Compute extrema depending on valid data proportion
    # Use default values if there is not enough valid data
    # Use percentiles otherwise
    try:
        ignored_mask, data_mask = build_masks(input_dir)
    except OnlyNightData:
        logger.warning('No day data in granule, the IDF would only contain '
                       'masked values. Exiting')
        raise EarlyExit()

    transforms.append(('contrast_from_pct',
                       {'targets': channels,
                        'valid_threshold': 0.001,  # 0.1%
                        'common_mask': ignored_mask,
                        'bands_mask': None,
                        'dynamic_bands_mask': get_bands_contrast_mask,
                        'min_percentile': 2.0,
                        'max_percentile': 99.99}))

    # Make sure max never exceeds the maximum sun glint
    transforms.append(('limit_extrema', {'targets': channels,
                                         'max_threshold': MAX_SUNGLINT}))

    # take default min and max if min is too high (lake, inland sea, clouds)
    #
    # At this point in the code, valid_min and valid_max have not been modified
    # so they are still the default values
    minmax_methods = {var_id: fix_extrema for var_id in channels}
    min_values = {var_id: v['valid_min'] for var_id, v in granule.vars.items()
                  if var_id in channels}
    max_values = {var_id: v['valid_max'] for var_id, v in granule.vars.items()
                  if var_id in channels}
    transforms.append(('extrema_methods', {'targets': channels,
                                           'methods': minmax_methods,
                                           'min_values': min_values,
                                           'max_values': max_values}))

    # Mask night data.
    transforms.append(('static_common_mask', {'targets': channels,
                                              'mask': data_mask}))

    # Share extrema among visible channels
    # TODO: remove this when SEAScope includes a way to configure channels
    # independently in RGB rendering
    transforms.append(('share_extrema', {'targets': channels}))

    # APPLY LOGARITHMIC SCALE
    # Pixels with a radiance equal or inferior to atmospheric correction
    # will be clipped to the minimal value as logscale cannot be applied to
    # negative values.
    transforms.append(('logscale', {'targets': channels,
                                    'base': 0}))

    output_opts['__export'] = channels

    yield input_opts, output_opts, granule, transforms



# vim: ts=4:sts=4:sw=4
#
# @date 2021-03-15
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.


""""""

import os
import numpy
import numpy.typing
import typing
import logging
import netCDF4
import scipy.ndimage
from scipy.interpolate import RegularGridInterpolator, griddata
from idf_converter.lib.types import InputOptions, OutputOptions, Granule
from idf_converter.lib.types import FormatterResult

logger = logging.getLogger(__name__)


def gaussian_filter_nan(array: numpy.typing.NDArray
                        ) -> numpy.typing.NDArray:
    """"""
    # Standard deviation for Gaussian kernel
    sigma = 8
    # Truncate filter at this many sigma
    truncate = sigma * 4
    # Set array with nan values to 0
    A0 = array.copy()
    array_invalid_ind = numpy.where(~numpy.isfinite(array))
    A0[array_invalid_ind] = 0
    A0g = scipy.ndimage.gaussian_filter(A0, sigma=sigma, truncate=truncate)
    # Build array with valid value set to 1, missing value to 0
    A1 = numpy.ones(A0.shape)
    A1[array_invalid_ind] = 0
    A1g = scipy.ndimage.gaussian_filter(A1, sigma=sigma, truncate=truncate)
    # Return ponderation of A0 by A1
    res: numpy.typing.NDArray = A0g / A1g
    res_invalid_mask = (~numpy.isfinite(res) | (res == 0))
    res_invalid_ind = numpy.where(res_invalid_mask)
    res[res_invalid_ind] = numpy.nan
    return res


def anomaly_from_clim(input_opts: InputOptions,
                      output_opts: OutputOptions,
                      granule: Granule,
                      *args: typing.Iterable[typing.Any],
                      climatology_path: str,
                      climatology_variable: str,
                      default_min: float,
                      default_max: float,
                      targets: typing.Tuple[str],
                      outputs: typing.Tuple[str],
                      extrapolate: bool,
                      interpolate: bool
                      ) -> FormatterResult:
    """"""
    start = granule.meta['time_coverage_start']
    stop = granule.meta['time_coverage_end']
    mid_date = start + (stop - start) / 2
    dday = mid_date.timetuple().tm_yday

    # Handle leap years case: offset everything by one day after 28th February
    if (0 == mid_date.timetuple().tm_year % 4) and (59 < dday):
        dday = dday - 1

    clim_file = climatology_path.replace('*', f'{dday:03d}')

    lon = granule.vars['lon']['array']
    lat = granule.vars['lat']['array']

    if not os.path.exists(clim_file):
        logger.warning('Anomaly not computed due to lack of climatology file')
        interp_clim_values = numpy.ma.masked_all(lon.shape)
    elif interpolate is False:
        if extrapolate is True:
            logger.warning('"extrapolate" option is ignored when the '
                           'climatology files are already on the same grid as '
                           'the input data, i.e. when the "interpolate" option'
                           ' is set to False')

        handler = netCDF4.Dataset(clim_file, 'r')
        interp_clim_values = handler[climatology_variable][0, :, :]
        handler.close()
    else:
        handler = netCDF4.Dataset(clim_file, 'r')
        clon = handler['lon'][:]
        clat = handler['lat'][:]
        cvalues = handler[climatology_variable][0, :, :]
        handler.close()
        clon2d, clat2d = numpy.meshgrid(clon, clat)

        if extrapolate is True:
            marray = numpy.ma.masked_invalid(cvalues)
            valid_ind = numpy.where(~numpy.ma.getmaskarray(marray))
            mlon = clon2d[valid_ind]
            mlat = clat2d[valid_ind]
            mcvalues = cvalues[valid_ind]
            cvalues = griddata((mlat, mlon), mcvalues.ravel(),
                               (clat2d, clon2d), method='nearest')

        func_interp = RegularGridInterpolator((clat, clon), cvalues,
                                              bounds_error=False,
                                              fill_value=None)
        if len(numpy.shape(lon)) == 1:
            lon, lat = numpy.meshgrid(granule.vars['lon']['array'],
                                      granule.vars['lat']['array'])

        interp_clim_values_1d = func_interp((lat.ravel(), lon.ravel()),
                                            method='linear')
        interp_clim_values = interp_clim_values_1d.reshape(lon.shape)

    for input_var, output_var in zip(targets, outputs):
        _input = granule.vars[input_var]['array']
        _input_mask_ind = numpy.where(numpy.isnan(_input))

        if output_var not in granule.vars.keys():
            granule.vars[output_var] = {}

        granule.vars[output_var]['name'] = output_var
        granule.vars[output_var]['options'] = {}
        granule.vars[output_var]['array'] = _input - interp_clim_values
        granule.vars[output_var]['array'][_input_mask_ind] = numpy.nan

        output_var_data = granule.vars[output_var]['array']
        output_var_mask = (numpy.ma.getmaskarray(output_var_data)
                           | ~numpy.isfinite(output_var_data))
        if 'valid_min' not in granule.vars[output_var].keys():
            vmin = default_min
            if not output_var_mask.all():
                vmin = max(default_min, numpy.nanmin(output_var_data))
            granule.vars[output_var]['valid_min'] = vmin

        if 'valid_max' not in granule.vars[output_var].keys():
            vmax = default_max
            if not output_var_mask.all():
                vmax = min(default_max, numpy.nanmax(output_var_data))
            granule.vars[output_var]['valid_max'] = vmax
    return (input_opts, output_opts, granule)



# -*- coding: utf-8 -*-
# vim: ts=4:sts=4:sw=4

import idf_converter


class TestPackage_Metadata(object):
    def test_has_version(self):
        assert hasattr(idf_converter, '__version__')

    def test_has_description(self):
        assert hasattr(idf_converter, '__description__')

    def test_has_author(self):
        assert hasattr(idf_converter, '__author__')

    def test_has_author_email(self):
        assert hasattr(idf_converter, '__author_email__')

    def test_has_url(self):
        assert hasattr(idf_converter, '__url__')

    def test_has_keywords(self):
        assert hasattr(idf_converter, '__keywords__')

    def test_has_classifiers(self):
        assert hasattr(idf_converter, '__classifiers__')



# vim: ts=4:sts=4:sw=4
#
# @date 2020-01-13
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import typing
import logging
from idf_converter.lib.types import InputOptions, OutputOptions, Granule
from idf_converter.lib.types import FormatterResult

logger = logging.getLogger(__name__)


class MissingVariableForRemoval(ValueError):
    """Error raised when a variable which has been listed as a target for
    removal is not available in the granule. Although the result is the same
    (i.e. the variable is not in the granule at the end of the operation), this
    should not happen."""
    def __init__(self, var_id: str) -> None:
        """"""
        self.var_id = var_id


def remove_vars(input_opts: InputOptions,
                output_opts: OutputOptions,
                granule: Granule,
                *args: typing.Iterable[typing.Any],
                targets: typing.Iterable[str]) -> FormatterResult:
    """"""
    for target in targets:
        _ = granule.vars.pop(target, None)
        if _ is None:
            raise MissingVariableForRemoval(target)

    output_opts['__export'] = [_ for _ in output_opts['__export']
                               if _ not in targets]
    return input_opts, output_opts, granule



# vim: ts=4:sts=4:sw=4
#
# @date 2019-09-25
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""This module defines the in-memory representation for granules as well as
helper methods to get submodules for a specific version of the IDF format.
"""

import copy
import numpy
import typing
import logging
import netCDF4
import datetime
import numpy.typing
import idf_converter.lib.time
from idf_converter.lib.types import InputOptions, OutputOptions, Granule
from idf_converter.lib.types import TransformsList, IdxSlices, IdxShifts

logger = logging.getLogger(__name__)


class EarlyExit(Exception):
    """Exception raised when the converter should exit before producing any IDF
    file (e.g. no valid data found in input file)."""
    pass


class IDFVersionNotSupported(ValueError):
    """Error raised when the IDF specifications are requested for an unknown
    version of the format."""
    def __init__(self, version: str) -> None:
        """"""
        self.version = version


class InvalidTimeCoverageOverride(ValueError):
    """Error raised when the global override for one of the time coverage
    bounds is provided in a format that the parsing method does not support."""
    def __init__(self, attr_name: str, attr_value: str) -> None:
        """"""
        self.attr_name = attr_name
        self.attr_value = attr_value


class MissingProcessingLevel(ValueError):
    """Error raised when neither the input file nor the input option contain
    the "processing_level" property which is required to build the granule
    identifier in case the input file contains several times."""
    pass


class MissingProductVersion(ValueError):
    """Error raised when neither the input file nor the input option contain
    the "product_version" property which is required to build the granule
    identifier in case the input file contains several times."""
    pass


class MissingFileVersion(ValueError):
    """Error raised when neither the input file nor the input option contain
    the "file_version" property which is required to build the granule
    identifier in case the input file contains several times."""
    pass


class MissingTransformName(Exception):
    """Error raised when the name of the transform that will be used by a
    transform-based option is not specified in the option definition (with the
    'using' key)."""
    pass


class MissingVirtualVariableTransformName(Exception):
    """Error raised when the name of the transform that will produce a virtual
    variable is not specified in the output options (with the "using" key)."""
    def __init__(self, virtual_var_id: str) -> None:
        """"""
        self.virtual_var_id = virtual_var_id


def get_idf_specs(version: str) -> typing.Any:
    """Get the specifications submodule for a specific version of the IDF
    format.

    Parameters
    ----------
    version: str
        Targetted IDF format version

    Returns
    -------
    module
        A module providing specifications for the requested IDF format version
    """
    if version in ('1.0',):
        import idf_converter.lib.idf.v1.specs
        return idf_converter.lib.idf.v1.specs
    raise IDFVersionNotSupported(version)


def get_idf_formatter(version: str) -> typing.Any:
    """Get the formatter submodule for a specific version of the IDF format.

    Parameters
    ----------
    version: str
        Targetted IDF format version

    Returns
    -------
    module
        A module providing formatting method for the requested IDF format
        version
    """
    if version in ('1.0',):
        import idf_converter.lib.idf.v1.fmt
        return idf_converter.lib.idf.v1.fmt
    raise IDFVersionNotSupported(version)


def create_granule(idf_version: str, data_model: str) -> Granule:
    """"""
    specs = get_idf_specs(idf_version)
    _data_model = data_model.lower()
    dimensions = [x for x in specs.SUPPORTED_MODELS[_data_model][:-2]
                  if x is not None]

    granule = Granule(idf_version, data_model, dimensions)
    return granule


def create_subgranule(original: Granule, output_opts: OutputOptions,
                      start_dt: datetime.datetime, stop_dt: datetime.datetime,
                      dtime: datetime.datetime) -> Granule:
    """"""
    collection = output_opts.get('collection')
    processing_level = original.meta.get('processing_level', None)
    if processing_level is None:
        logger.error('Missing "processing_level" global attribute')
        raise MissingProcessingLevel()
    product_version = original.meta.get('product_version', None)
    if product_version is None:
        logger.error('Missing "product_version" global attribute')
        raise MissingProductVersion()
    file_version = original.meta.get('file_version', None)
    if file_version is None:
        logger.error('Missing "file_version" global attribute')
        raise MissingFileVersion()

    # <ProductString>_<YYYYMMDD><HHMMSS>_<ProcessingLevel>_v<ProductVersion>_fv<FileVersion>  # noqa
    indicative_datetime = f'{dtime:%Y%m%d%H%M%S}'
    granule_name = f'{collection}_{indicative_datetime}'
    granule_name = f'{granule_name}_{processing_level}'
    granule_name = f'{granule_name}_v{product_version}'
    granule_name = f'{granule_name}_fv{file_version}'

    subgranule: Granule = create_granule(original.idf_version,
                                         original.data_model)
    subgranule.meta = copy.deepcopy(original.meta)

    subgranule.dims = copy.deepcopy(original.dims)

    subgranule.meta['idf_granule_id'] = granule_name
    subgranule.meta['time_coverage_start'] = start_dt
    subgranule.meta['time_coverage_end'] = stop_dt

    return subgranule


def extract_global_attributes(f_handler: netCDF4.Dataset,
                              input_opts: InputOptions,
                              granule: Granule) -> None:
    """"""
    _parse_datetime = idf_converter.lib.time.parse_datetime
    specs = get_idf_specs(granule.idf_version)
    for attr_name in specs.MANDATORY_ATTRIBUTES:
        if attr_name in granule.meta.keys():
            continue
        if hasattr(f_handler, attr_name):
            attr_value = getattr(f_handler, attr_name)
            is_time_coverage_attr = (attr_name in ('time_coverage_start',
                                                   'time_coverage_end'))
            if is_time_coverage_attr and isinstance(attr_value, str):
                suffix_attr = f'{attr_name}_suffix'
                suffix = input_opts.get(suffix_attr, '')
                attr_value = f'{attr_value}{suffix}'
                time_coverage_fmt = input_opts.get('time_coverage_format',
                                                   None)
                try:
                    attr_value = _parse_datetime(attr_value, time_coverage_fmt)
                except idf_converter.lib.time.InvalidDatetimeFormat:
                    logger.warning(f'{attr_name} found in input file but the '
                                   'datetime format is not supported and the '
                                   'attribute has therefore been ignored')
                    continue
            granule.meta[attr_name] = attr_value

    for attr_name in specs.SUPPORTED_ATTRIBUTES:
        if attr_name in granule.meta.keys():
            continue
        if hasattr(f_handler, attr_name):
            granule.meta[attr_name] = getattr(f_handler, attr_name)

    for attr_name in specs.CF_ATTRIBUTES:
        if attr_name in granule.meta.keys():
            continue
        if hasattr(f_handler, attr_name):
            granule.meta[attr_name] = getattr(f_handler, attr_name)

    for attr_name in specs.OPTIONAL_ATTRIBUTES:
        if attr_name in granule.meta.keys():
            continue
        if hasattr(f_handler, attr_name):
            granule.meta[attr_name] = getattr(f_handler, attr_name)

    for attr_name in specs.SPECIAL_ATTRIBUTES:
        if attr_name in granule.meta.keys():
            continue
        if hasattr(f_handler, attr_name):
            granule.meta[attr_name] = getattr(f_handler, attr_name)


def extract_variable_attributes(f_handler: netCDF4.Dataset,
                                var_id: str,
                                granule: Granule) -> None:
    """"""
    extracted_attrs = []
    specs = get_idf_specs(granule.idf_version)

    if var_id not in granule.vars.keys():
        granule.vars[var_id] = {'options': {}}

    if var_id not in f_handler.variables.keys():
        return

    src_var = f_handler.variables[var_id]
    for attr_name in specs.VARIABLE_MANDATORY_ATTRIBUTES:
        if attr_name in granule.vars[var_id].keys():
            continue
        if hasattr(src_var, attr_name):
            granule.vars[var_id][attr_name] = getattr(src_var, attr_name)
            extracted_attrs.append(attr_name)

    for attr_name in specs.VARIABLE_CF_ATTRIBUTES:
        if attr_name in granule.vars[var_id].keys():
            continue
        if hasattr(src_var, attr_name):
            granule.vars[var_id][attr_name] = getattr(src_var, attr_name)
            extracted_attrs.append(attr_name)

    for attr_name in specs.VARIABLE_OPTIONAL_ATTRIBUTES:
        if attr_name in granule.vars[var_id].keys():
            continue
        if hasattr(src_var, attr_name):
            granule.vars[var_id][attr_name] = getattr(src_var, attr_name)
            extracted_attrs.append(attr_name)

    # According to the NetCDF User Guide, valid_min and valid_max must be
    # expressed as packed values if there is an offset and a scale factor
    # but the IDF converter needs the unpacked values
    is_packed = (('add_offset' in extracted_attrs) and
                 ('scale_factor' in extracted_attrs))
    if is_packed and ('valid_min' in extracted_attrs):
        # Unpack valid_min
        offset = granule.vars[var_id]['add_offset']
        scale = granule.vars[var_id]['scale_factor']
        unpacked_value = offset + scale * granule.vars[var_id]['valid_min']
        granule.vars[var_id]['valid_min'] = unpacked_value
    if is_packed and ('valid_max' in extracted_attrs):
        # Unpack valid_max
        offset = granule.vars[var_id]['add_offset']
        scale = granule.vars[var_id]['scale_factor']
        unpacked_value = offset + scale * granule.vars[var_id]['valid_max']
        granule.vars[var_id]['valid_max'] = unpacked_value

    # Values extracted from the input file are unpacked, i.e. add_offset and
    # scale_factor have already been applied. Remove them from the variable
    # attributes to avoid confusion
    if 'scale_factor' in granule.vars[var_id].keys():
        del granule.vars[var_id]['scale_factor']
    if 'add_offset' in granule.vars[var_id].keys():
        del granule.vars[var_id]['add_offset']


def _extract_values(f_handler: netCDF4.Dataset,
                    var_id: str,
                    idx: IdxSlices,
                    value_placeholder: typing.Any,
                    mask_state: typing.Optional[bool]
                    ) -> typing.Tuple[numpy.typing.ArrayLike,
                                      typing.Optional[bool]]:
    """"""
    _idx = tuple(idx)
    band: numpy.typing.ArrayLike  # simply to declare type for mypy

    # When there is a placeholder value and automasking is active (default)
    # netCDF4 replaces missing/invalid values by the placeholder, but if
    # the placeholder value is NaN this behavior triggers Numpy warnings
    # Note: netCDF4 uses missing_value as placeholder if this attribute is
    # provided, otherwise it uses _FillValue as a fallback.
    if (value_placeholder is not None) and numpy.isnan(value_placeholder):
        f_handler.variables[var_id].set_auto_mask(False)
        band = f_handler.variables[var_id][_idx]
        band = numpy.ma.masked_invalid(band)
        if mask_state is None:
            mask_state = True
    else:
        band = f_handler.variables[var_id][_idx]

    return band, mask_state


def _extract_shifted_values(f_handler: netCDF4.Dataset,
                            var_id: str,
                            idx: IdxSlices,
                            value_placeholder: typing.Any,
                            mask_state: typing.Optional[bool],
                            shift: IdxShifts
                            ) -> typing.Tuple[numpy.typing.ArrayLike,
                                              typing.Optional[bool]]:
    """"""
    _idx = tuple(idx)
    band: numpy.typing.ArrayLike  # simply to declare type for mypy

    # When there is a placeholder value and automasking is active (default)
    # netCDF4 replaces missing/invalid values by the placeholder, but if
    # the placeholder value is NaN this behavior triggers Numpy warnings
    # Note: netCDF4 uses missing_value as placeholder if this attribute is
    # provided, otherwise it uses _FillValue as a fallback.
    if (value_placeholder is not None) and numpy.isnan(value_placeholder):
        f_handler.variables[var_id].set_auto_mask(False)
        band = f_handler.variables[var_id]
        band = numpy.ma.masked_invalid(band)
        if mask_state is None:
            mask_state = True
    else:
        band = f_handler.variables[var_id]

    for axis, amount in shift.items():
        band = numpy.roll(band, amount, axis=axis)

    return band[_idx], mask_state


def extract_variable_values(f_handler: netCDF4.Dataset,
                            var_id: str,
                            mask_state: typing.Optional[bool] = None,
                            idx: typing.Optional[IdxSlices] = None,
                            shifts: typing.Optional[IdxShifts] = None
                            ) -> numpy.typing.ArrayLike:
    """"""
    if idx is None:
        idx = [slice(None, None, None)]

    # Detected whether variable has a placeholder value for missing or
    # invalid data
    has_fill_value = hasattr(f_handler.variables[var_id], '_FillValue')
    has_missing_value = hasattr(f_handler.variables[var_id], 'missing_value')

    value_placeholder = None
    if has_missing_value:
        value_placeholder = f_handler.variables[var_id].missing_value
    elif has_fill_value:
        value_placeholder = f_handler.variables[var_id]._FillValue

    # Some netCDF files use a missing_value or a _FillValue that is not
    # compatible with the data type of the variable. The data producers should
    # fix their files because they are invalid, but it is unlikely so the
    # idf_converter considers the invalid value to be a NaN.
    if value_placeholder is not None:
        dtype = f_handler.variables[var_id].datatype
        if not numpy.can_cast(value_placeholder, dtype):
            logger.warning(f'The placeholder value {value_placeholder} for '
                           f'"{var_id}" does not match its data type')
            value_placeholder = float('nan')

    if mask_state is not None:
        f_handler.variables[var_id].set_auto_mask(mask_state)

    if (shifts is None) or (0 >= len(list(shifts.keys()))):
        band, mask_state = _extract_values(f_handler, var_id, idx,
                                           value_placeholder, mask_state)
    else:
        band, mask_state = _extract_shifted_values(f_handler, var_id, idx,
                                                   value_placeholder,
                                                   mask_state, shifts)
    if mask_state is not None:
        f_handler.variables[var_id].set_auto_mask(mask_state)

    return band


def find_whitespace(str_buffer: str) -> int:
    """"""
    return next((i for i, s in enumerate(str_buffer) if s.isspace()), -1)


def rfind_whitespace(str_buffer: str) -> int:
    """"""
    max_index = len(str_buffer) - 1
    reversed_buffer = reversed(str_buffer)
    return next((max_index - i for i, s in enumerate(reversed_buffer)
                 if s.isspace()),
                -1)


def get_next_pair(str_buffer: str, original: str, kv_separator: str,
                  offset: int) -> typing.Tuple[typing.Optional[str],
                                               typing.Optional[str],
                                               int]:
    """"""
    str_len = len(str_buffer)
    if 0 >= str_len:
        return (None, None, offset)

    i_equal = str_buffer.find(kv_separator, offset)

    # Find key
    tmp_buffer = str_buffer[0:i_equal].rstrip()
    i_key = 1 + rfind_whitespace(tmp_buffer)
    key = original[i_key:i_equal].strip()

    # Find value
    i_value_start = str_len - len(str_buffer[(1 + i_equal):].lstrip())
    first_char = str_buffer[i_value_start]
    if first_char in ('"', "'"):
        i_value_stop = str_buffer.find(first_char, 1 + i_value_start)
        if 0 > i_value_stop:
            i_value_stop = str_len
        value = original[i_value_start:i_value_stop]
        i_value_stop = min(1 + i_value_stop, str_len)
        value = value[1:]  # ending quote is already gone but starting remains
    else:
        tmp_buffer = str_buffer[i_value_start:]
        i_next_equal = str_buffer.find(kv_separator, i_value_start)
        if 0 > i_next_equal:
            i_value_stop = str_len
        else:
            tmp_buffer = str_buffer[0:i_next_equal].rstrip()
            i_value_stop = rfind_whitespace(tmp_buffer)
        value = original[i_value_start:i_value_stop].strip()
    new_offset = i_value_stop
    return (key, value, new_offset)


def parse_opt(text: str, pair_separator: str,
              kv_separator: str) -> typing.Iterator[typing.Tuple[str, str]]:
    """"""
    # Remove spaces on both ends of the string
    stripped_text = text.strip()

    if ' ' != pair_separator:
        stripped_text = stripped_text.replace(pair_separator, ' ')

    offset = 0
    str_len = len(stripped_text)
    while offset < str_len - 1:
        key, value, offset = get_next_pair(stripped_text, text,
                                           kv_separator, offset)
        if (key is not None) and (value is not None):
            yield (key, value)


def apply_global_overrides(input_opts: InputOptions, granule: Granule) -> None:
    """
    Global overrides are provided as a comma-separated list of key:value.
    No space is allowed around the separating comma.
    No trailing comma.
    Values cannot contain an equal sign.
    """
    specs = get_idf_specs(granule.idf_version)
    attrs_str = input_opts.get('global_overrides', None)
    if attrs_str is None:
        return
    attrs = parse_opt(attrs_str, ',', ':')
    for _attr_name, _attr_value in attrs:
        attr_name = _attr_name.strip()
        attr_value = _attr_value.strip()
        is_time_coverage_attr = (attr_name in ('time_coverage_start',
                                               'time_coverage_end'))
        if is_time_coverage_attr and isinstance(attr_value, str):
            _parse_datetime = idf_converter.lib.time.parse_datetime
            time_coverage_fmt = input_opts.get('time_coverage_format', None)
            try:
                dt_value = _parse_datetime(attr_value, time_coverage_fmt)
            except idf_converter.lib.time.InvalidDatetimeFormat:
                logger.warning(f'{attr_name} found in input file but the '
                               'datetime format is not supported and the '
                               'attribute has therefore been ignored')
                raise InvalidTimeCoverageOverride(attr_name, attr_value)
            granule.meta[attr_name] = dt_value
        elif attr_name in specs.NUMERICAL_ATTRIBUTES:
            granule.meta[attr_name] = float(attr_value)
        else:
            granule.meta[attr_name] = attr_value


def apply_var_overrides(input_opts: InputOptions, granule: Granule) -> None:
    """"""
    specs = get_idf_specs(granule.idf_version)

    for override_key in input_opts.keys():
        if not override_key.startswith('override('):
            continue

        var_id = override_key.strip()[9:-1]
        attrs_str = input_opts[override_key]
        attrs = parse_opt(attrs_str, ',', ':')
        for _attr_name, _attr_value in attrs:
            attr_name = _attr_name.strip()
            attr_value = _attr_value.strip()
            if var_id not in granule.vars.keys():
                granule.vars[var_id] = {'options': {}}

            if attr_name in specs.VARIABLE_NUMERICAL_ATTRIBUTES:
                auto_attr = attr_name in ('valid_min', 'valid_max')
                if (auto_attr is True) and ('auto' == attr_value):
                    granule.vars[var_id][attr_name] = attr_value
                else:
                    granule.vars[var_id][attr_name] = float(attr_value)
            else:
                granule.vars[var_id][attr_name] = attr_value


def build_flag_mask(var_id: str,
                    band: numpy.typing.NDArray,
                    flag_band: numpy.typing.NDArray,
                    granule: Granule,
                    ) -> numpy.typing.NDArray:

    """"""
    flag_var_id = granule.vars[var_id].pop('flag_variable', None)
    flag_min_value = granule.vars[var_id].pop('flag_min', None)
    flag_max_value = granule.vars[var_id].pop('flag_max', None)

    has_flag_var = (flag_var_id is not None)
    has_flag_criteria = ((flag_min_value is not None) or
                         (flag_max_value is not None))
    if (has_flag_var and has_flag_criteria) is False:
        band_shape = numpy.shape(band)
        default_mask: numpy.typing.NDArray
        default_mask = numpy.ma.make_mask_none(band_shape)
        return default_mask

    _qual: numpy.typing.NDArray
    _qual = numpy.ma.masked_array(flag_band)

    # _mask = True when the data should be considered invalid, i.e. when:

    # - the flag variable is masked (i.e. no information on data quality)
    _mask: numpy.typing.NDArray = numpy.ma.getmaskarray(_qual)
    _qual_type = numpy.obj2sctype(_qual.dtype)

    # - the value of the flag variable is strictly below the lower theshold
    if flag_min_value is not None:
        _mask = (_mask | (_qual < _qual_type(flag_min_value)))

    # - the value of the flag variable is strictly above the upper threshold
    if flag_max_value is not None:
        _mask = (_mask | (_qual > _qual_type(flag_max_value)))

    return _mask


def parse_transform_option(text: str
                           ) -> typing.Tuple[str,
                                             typing.Dict[str, typing.Any]]:
    """"""
    _transform_attrs = idf_converter.lib.parse_opt(text, ',', ':')
    transform_attrs: typing.Dict[str, typing.Any] = {}
    for name, value in _transform_attrs:
        if ';' in value:
            transform_attrs[name] = [_.strip()
                                     for _ in value.split(';')
                                     if 0 < len(_.strip())]
        else:
            try:
                transform_attrs[name] = float(value.strip())
            except ValueError:
                _value = value.strip()
                if _value.lower() in ('yes', 'true'):
                    transform_attrs[name] = True
                elif _value.lower() in ('no', 'false'):
                    transform_attrs[name] = False
                else:
                    transform_attrs[name] = _value

    if 'using' not in transform_attrs.keys():
        logger.debug(text)
        raise MissingTransformName()

    transform_name = transform_attrs.pop('using')
    return (transform_name, transform_attrs)


def add_virtual_variables(output_opts: OutputOptions,
                          transforms: TransformsList
                          ) -> None:
    """"""
    virtual_var_ids = []
    for key in output_opts.keys():
        if key.startswith('add_variables('):
            var_ids_str = key[14:-1]
            var_ids = [_.strip() for _ in var_ids_str.split(',')
                       if 0 < len(_.strip())]
            var_cfg = output_opts[key]
            try:
                transform_option = parse_transform_option(var_cfg)
            except MissingTransformName:
                raise MissingVirtualVariableTransformName(var_ids_str)
            transforms.append(transform_option)
            virtual_var_ids.extend(var_ids)

    exported_ids = output_opts.get('__export', [])
    exported_ids.extend(virtual_var_ids)

    # Make sure there are no duplicates
    exported_ids = list(set(exported_ids))

    output_opts['__export'] = exported_ids


def remove_variables(output_opts: OutputOptions,
                     transforms: TransformsList
                     ) -> None:
    """"""
    if 'remove_variables' in output_opts.keys():
        var_ids = [_.strip()
                   for _ in output_opts['remove_variables'].split(',')]
        if 0 < len(var_ids):
            transforms.append(('remove_vars', {'targets': var_ids}))


def _safe_mod_angle(num: numpy.typing.NDArray,
                    denom: float,
                    decimals: int
                    ) -> typing.Tuple[numpy.typing.NDArray, bool]:
    """"""
    # The difference between angles (or longitudes) values read from the input
    # file might be very small values: having such values might cause underflow
    # errors when calling the numpy implementation of modulo.
    # As a workaround, since we know that angles/longitudes are not meant to
    # reach very high values we can multiply both operands of the modulo
    # function by a fixed factor > 1 to avoid the underflows and then divide
    # the result by the same factor to get the actual result.
    # The uflow_fix variable will play the role of the fixed factor in this
    # method.

    values: numpy.typing.NDArray
    values_truncated_to_0 = False
    try:
        values = numpy.mod(num, denom)
    except FloatingPointError:
        uflow_fix = 10 ** decimals

        _rounded_num = numpy.round(num, decimals=decimals)
        _scaled_num = uflow_fix * _rounded_num
        _scaled_denom = uflow_fix * denom

        mask = (numpy.abs(_scaled_num) < 1.0)  # i.e. |angle| < 10^-decimals
        if mask.any():
            _scaled_num[numpy.where(mask)] = 0.0
            values_truncated_to_0 = True

        _values = numpy.mod(_scaled_num,
                            _scaled_denom,
                            where=numpy.isfinite(_scaled_num))

        # Make sure that dividing by uflow_fix will not trigger an underflow
        eps = numpy.finfo(num.dtype).eps
        underflow_mask = (numpy.abs(_values) < uflow_fix * eps)
        if underflow_mask.any():
            _values[numpy.where(underflow_mask)] = 0.0
            values_truncated_to_0 = True

        values = _values / uflow_fix
    return values, values_truncated_to_0


def mod360(num: numpy.typing.NDArray,
           decimals: int
           ) -> typing.Tuple[numpy.typing.NDArray, bool]:
    """"""
    return _safe_mod_angle(num, 360.0, decimals)


def mod2pi(num: numpy.typing.NDArray,
           decimals: int
           ) -> typing.Tuple[numpy.typing.NDArray, bool]:
    """"""
    return _safe_mod_angle(num, 2 * numpy.pi, decimals)



# vim: ts=4:sts=4:sw=4
#
# @date 2019-09-25
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

import numpy
import numpy.typing
import typing
import logging
import dateutil
import datetime

logger = logging.getLogger(__name__)


class Granule(object):
    """"""
    def __init__(self, idf_version: str, data_model: str,
                 dimensions: typing.List[str]) -> None:
        """"""
        self.idf_version: str = idf_version
        self.vars: typing.Dict[str, typing.Any] = {}
        self.dims: typing.Dict[str, int] = {}
        self.meta: typing.Dict[str, typing.Any] = {}
        self.data_model: str = data_model
        self._data_model_dimensions = dimensions
        for dim in dimensions:
            self.dims[dim] = 0

    def get_data_model_dimensions(self) -> typing.List[str]:
        """"""
        return self._data_model_dimensions.copy()

    def validate(self) -> bool:
        """"""
        return False


InputOptions = typing.Dict[str, typing.Any]
OutputOptions = typing.Dict[str, typing.Any]
FormatterJob = typing.Tuple[str, typing.Dict[str, typing.Any]]
AtmCorrData = typing.Dict[str, typing.Any]
NumValuesDict = typing.Dict[str, numpy.typing.NDArray]
Extrema = typing.Tuple[float, float]
Duration = dateutil.relativedelta.relativedelta
RelativeBound = typing.Union[str, float]
CoverageInfo = typing.Tuple[typing.Optional[datetime.datetime],
                            typing.Optional[datetime.datetime],
                            typing.Optional[RelativeBound],
                            typing.Optional[RelativeBound]]
Coverage = typing.Tuple[typing.Optional[datetime.datetime],
                        typing.Optional[datetime.datetime]]
ReaderResult = typing.Tuple[typing.Dict[str, typing.Any],
                            typing.Dict[str, typing.Any],
                            Granule,
                            typing.List[FormatterJob]]
FormatterResult = typing.Tuple[InputOptions, OutputOptions, Granule]
ProcessedLists = typing.Tuple[typing.Iterable[str],
                              typing.Iterable[str],
                              typing.Iterable[str]]
TransformsList = typing.List[typing.Tuple[str, typing.Dict[str, typing.Any]]]
IdxSlices = typing.List[slice]
IdxShifts = typing.Dict[int, int]
GCPs = typing.Tuple[numpy.typing.NDArray, numpy.typing.NDArray,
                    numpy.typing.NDArray, numpy.typing.NDArray]
